"""
Client for connecting to Blackboard.
"""
import base64
import copy
import json
import logging
import time
from http import HTTPStatus
from urllib.parse import urljoin

import requests

from django.apps import apps
from django.db import transaction

from integrated_channels.blackboard.exporters.content_metadata import BLACKBOARD_COURSE_CONTENT_NAME
from integrated_channels.exceptions import ClientError
from integrated_channels.integrated_channel.client import IntegratedChannelApiClient
from integrated_channels.utils import generate_formatted_log, refresh_session_if_expired, stringify_and_store_api_record

LOGGER = logging.getLogger(__name__)

# TODO: Refactor candidate (duplication with canvas client)
GRADEBOOK_PATH = '/learn/api/public/v1/courses/{course_id}/gradebook/columns'
ENROLLMENT_PATH = '/learn/api/public/v1/courses/{course_id}/users'
COURSE_PATH = '/learn/api/public/v1/courses'
POST_GRADE_COLUMN_PATH = '/learn/api/public/v2/courses/{course_id}/gradebook/columns'
PATCH_GRADE_COLUMN_PATH = '/learn/api/public/v2/courses/{course_id}/gradebook/columns/{column_id}'
POST_GRADE_PATH = '/learn/api/public/v2/courses/{course_id}/gradebook/columns/{column_id}/users/{user_id}'
COURSE_V3_PATH = '/learn/api/public/v3/courses/{course_id}'
COURSES_V3_PATH = '/learn/api/public/v3/courses'
COURSE_CONTENT_PATH = '/learn/api/public/v1/courses/{course_id}/contents'
COURSE_CONTENT_CHILDREN_PATH = '/learn/api/public/v1/courses/{course_id}/contents/{content_id}/children'
COURSE_CONTENT_DELETE_PATH = '/learn/api/public/v1/courses/{course_id}/contents/{content_id}'
GRADEBOOK_COLUMN_DESC = "edX learner's grade"
PAGE_TRAVERSAL_LIMIT = 250


class BlackboardAPIClient(IntegratedChannelApiClient):
    """
    Client for connecting to Blackboard.
    """

    def __init__(self, enterprise_configuration):
        """
        Instantiate a new client.

        Args:
            enterprise_configuration (BlackboardEnterpriseCustomerConfiguration): An enterprise customers's
            configuration model for connecting with Blackboard
        """
        super().__init__(enterprise_configuration)
        BlackboardGlobalConfiguration = apps.get_model(
            'blackboard',
            'BlackboardGlobalConfiguration'
        )
        self.global_blackboard_config = BlackboardGlobalConfiguration.current()
        self.config = apps.get_app_config('blackboard')
        self.session = None
        self.expires_at = None

    def create_content_metadata(self, serialized_data):
        """
        Create a course from serialized course metadata
        Returns: (int, str) Status code, Status message
        """
        channel_metadata_item = json.loads(serialized_data.decode('utf-8'))
        BlackboardAPIClient._validate_channel_metadata(channel_metadata_item)
        BlackboardAPIClient._validate_full_course_metadata(channel_metadata_item)

        external_id = channel_metadata_item.get('externalId')

        # blackboard does not support all characters in our courseIds so let's gen a hash instead
        course_id_generated = self.generate_blackboard_course_id(external_id)

        copy_of_channel_metadata = copy.deepcopy(channel_metadata_item)
        copy_of_channel_metadata['course_metadata']['courseId'] = course_id_generated

        self._create_session()
        create_url = self.generate_course_create_url()
        try:
            response = self._post(create_url, copy_of_channel_metadata['course_metadata'])
        except ClientError as error:
            if error.status_code == 409 and 'Unique ID conflicts' in error.message:
                # course already exists!
                msg_body = (f"Course already exists with course_id {external_id},"
                            f" and generated course_id: {course_id_generated}, not attempting creation")
                LOGGER.warning(generate_formatted_log(
                    self.enterprise_configuration.channel_code(),
                    self.enterprise_configuration.enterprise_customer.uuid,
                    None,
                    external_id,
                    msg_body,
                ))
                return HTTPStatus.NOT_MODIFIED.value, msg_body
            else:
                raise error

        # We wrap error handling in the post, but sanity check for the ID
        bb_course_id = response.json().get('id')
        if not bb_course_id:
            raise ClientError(
                'Something went wrong while creating base course object on Blackboard. Could not retrieve course ID.',
                HTTPStatus.NOT_FOUND.value
            )

        course_created_response = self.create_integration_content_for_course(bb_course_id, copy_of_channel_metadata)

        success_body = 'Successfully created Blackboard integration course={bb_course_id} with integration ' \
                       'content={bb_child_id}'.format(
                           bb_course_id=bb_course_id,
                           bb_child_id=course_created_response.json().get('id'),
                       )

        return course_created_response.status_code, success_body

    def update_content_metadata(self, serialized_data):
        """Apply changes to a course if applicable"""
        self._create_session()
        channel_metadata_item = json.loads(serialized_data.decode('utf-8'))

        BlackboardAPIClient._validate_channel_metadata(channel_metadata_item)
        external_id = channel_metadata_item.get('externalId')
        course_id = self._resolve_blackboard_course_id(external_id)
        BlackboardAPIClient._validate_course_id(course_id, external_id)

        update_url = self.generate_course_update_url(course_id)
        response = self._patch(update_url, channel_metadata_item.get('course_metadata'))

        bb_course_id = response.json().get('id')

        if not bb_course_id:
            raise ClientError(
                'Unable to update course={} content on Blackboard. Failed to retrieve ID from course update response:'
                '(status_code={}, body={})'.format(external_id, response.status_code, response.text)
            )

        course_contents_url = self.generate_create_course_content_url(bb_course_id)
        course_contents_response = self._get(course_contents_url)

        bb_contents = course_contents_response.json().get('results')
        bb_content_id = None

        for content in bb_contents:
            if content.get('title') == BLACKBOARD_COURSE_CONTENT_NAME:
                bb_content_id = content.get('id')

        if not bb_content_id:
            LOGGER.info(
                generate_formatted_log(
                    self.enterprise_configuration.channel_code(),
                    self.enterprise_configuration.enterprise_customer.uuid,
                    None,
                    course_id,
                    'Blackboard integrated course content not found. Generating content.'
                )
            )
        else:
            course_content_delete_url = self.generate_course_content_delete_url(bb_course_id, bb_content_id)
            self._delete(course_content_delete_url)

        course_updated_response = self.update_integration_content_for_course(
            bb_course_id,
            channel_metadata_item,
        )

        return course_updated_response.status_code, course_updated_response.text

    def delete_content_metadata(self, serialized_data):
        """Delete course from blackboard (performs full delete as of now)"""
        self._create_session()
        channel_metadata_item = json.loads(serialized_data.decode("utf-8"))

        BlackboardAPIClient._validate_channel_metadata(channel_metadata_item)
        external_id = channel_metadata_item.get('externalId')
        course_id = self._resolve_blackboard_course_id(external_id)

        if not course_id:
            return HTTPStatus.OK.value, 'Course:{} already removed.'.format(external_id)

        LOGGER.info(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                course_id,
                f'Deleting course with courseId: {course_id}'
            )
        )
        update_url = self.generate_course_update_url(course_id)
        response = self._delete(update_url)
        return response.status_code, response.text

    def create_assessment_reporting(self, user_id, payload):
        """
        Post a learner's subsection assessment grade to the integrated Blackboard course.

        Parameters:
        -----------
            user_id (str): The shared email between a user's edX account and Blackboard account
            payload (str): The (string representation) of the learner data information
        """
        self._create_session()
        learner_data = json.loads(payload)
        external_id = learner_data.get('courseID')

        course_id = self._resolve_blackboard_course_id(external_id)
        BlackboardAPIClient._validate_course_id(course_id, external_id)

        blackboard_user_id = self._get_bb_user_id_from_enrollments(user_id, course_id)
        grade_column_id = self._get_or_create_integrated_grade_column(
            course_id,
            learner_data.get('subsection_name'),
            learner_data.get('subsectionID'),
            learner_data.get('points_possible'),
        )

        grade = learner_data.get('points_earned')
        submission_response = self._submit_grade_to_blackboard(grade, course_id, grade_column_id, blackboard_user_id)

        success_body = 'Successfully posted grade of {grade} to course:{course_id} for user:{user_email}.'.format(
            grade=grade,
            course_id=external_id,
            user_email=user_id,
        )
        return submission_response.status_code, success_body

    def create_course_completion(self, user_id, payload):
        """
        Post a final course grade to the integrated Blackboard course.

        Parameters:
        -----------
            user_id (str): The shared email between a user's edX account and Blackboard account
            payload (str): The (string representation) of the learner data information

        Example payload:
        ---------------
            '{
                courseID: course-edx+555+3T2020,
                score: 0.85,
                completedTimestamp: 1602265162589,
            }'

        """
        self._create_session()
        learner_data = json.loads(payload)
        external_id = learner_data.get('courseID')

        course_id = self._resolve_blackboard_course_id(external_id)
        BlackboardAPIClient._validate_course_id(course_id, external_id)

        blackboard_user_id = self._get_bb_user_id_from_enrollments(user_id, course_id)
        grade_column_id = self._get_or_create_integrated_grade_column(
            course_id,
            "(edX Integration) Final Grade",
            "edx_final_grade",
            include_in_calculations=True,
        )

        grade = learner_data.get('grade') * 100
        submission_response = self._submit_grade_to_blackboard(grade, course_id, grade_column_id, blackboard_user_id)

        success_body = 'Successfully posted grade of {grade} to course:{course_id} for user:{user_email}.'.format(
            grade=grade,
            course_id=external_id,
            user_email=user_id,
        )
        return submission_response.status_code, success_body

    def delete_course_completion(self, user_id, payload):
        """TODO: course completion deletion is currently not easily supported"""

    @staticmethod
    def _validate_channel_metadata(channel_metadata_item):
        """
        Raise error if external_id invalid or not found
        """
        if 'externalId' not in channel_metadata_item:
            raise ClientError("No externalId found in metadata, please check json data format", 400)

    @staticmethod
    def _validate_course_id(course_id, external_id):
        """
        Raise error if course_id invalid
        """
        if not course_id:
            raise ClientError(
                'Could not find course:{} on Blackboard'.format(external_id),
                HTTPStatus.NOT_FOUND.value
            )

    @staticmethod
    def _validate_full_course_metadata(channel_metadata_items):
        """
        Raise error if course_metadata, course_content_metadata or course_child_content_metadata are invalid.
        """
        metadata_set = set(channel_metadata_items.keys())
        course_metadata_groups = {'course_metadata', 'course_content_metadata', 'course_child_content_metadata'}
        if not course_metadata_groups.issubset(metadata_set):
            raise ClientError(
                'Could not find course metadata group(s):{} necessary to create Blackboard integrated course.'.format(
                    course_metadata_groups - metadata_set
                ),
                HTTPStatus.NOT_FOUND.value
            )

    def _create_session(self):
        """
        Will only create a new session if token expiry has been reached
        """
        self.session, self.expires_at = refresh_session_if_expired(
            self._get_oauth_access_token,
            self.session,
            self.expires_at,
        )

    def _formatted_message(self, msg):
        return generate_formatted_log(
            self.enterprise_configuration.channel_code(),
            self.enterprise_configuration.enterprise_customer.uuid,
            None,
            None,
            msg,
        )

    def _log_info(self, msg):
        LOGGER.info(self._formatted_message(msg))

    def _log_error(self, msg):
        LOGGER.error(self._formatted_message(msg))

    def _get_oauth_access_token(self):
        """Fetch access token using refresh_token workflow from Blackboard

        Using atomic block for refresh token handling code, because we need to use the
        most recently obtained refresh token always
        Since any time we use a refresh token to get a new one, the prior one is invalidated
        and we MUST use the new one for the next request.

        Returns:
            access_token (str): the OAuth access token to access the Blackboard server
            expires_in (int): the number of seconds after which token will expire
        Raises:
            HTTPError: If we received a failure response code.
            ClientError: If an unexpected response format was received that we could not parse.
        """

        if (not self.enterprise_configuration.blackboard_base_url
                or not self.config.oauth_token_auth_path):
            raise ClientError(
                "Failed to generate oauth access token: oauth path missing from configuration.",
                HTTPStatus.INTERNAL_SERVER_ERROR.value
            )
        auth_token_url = urljoin(
            self.enterprise_configuration.blackboard_base_url,
            self.config.oauth_token_auth_path,
        )

        # Refresh token handling atomic block
        # DO NOT USE self.enterprise_config in this block!
        # Here we don't use the config passed from the constructor directly
        # instead we fetch a transaction safe instance using select_for_update
        # so that multiple sessions such as this
        # don't interfere with each other's Blackboard refresh token state. Blackboard only allows
        # only valid refresh token at a time, and once it's used it's invalidated and needs to be
        # replaced by a new token
        with transaction.atomic():
            channel_config = apps.get_model(
                'blackboard',
                'BlackboardEnterpriseCustomerConfiguration'
            ).objects.select_for_update().get(pk=self.enterprise_configuration.pk)
            if not channel_config.refresh_token:
                raise ClientError(
                    "Failed to generate oauth access token: Refresh token required.",
                    HTTPStatus.INTERNAL_SERVER_ERROR.value
                )

            auth_token_params = {
                'grant_type': 'refresh_token',
                'refresh_token': channel_config.refresh_token,
            }

            auth_response = requests.post(
                auth_token_url,
                auth_token_params,
                headers={
                    'Authorization': self._create_auth_header(),
                    'Content-Type': 'application/x-www-form-urlencoded'
                }
            )
            if auth_response.status_code >= 400:
                raise ClientError(
                    f"BLACKBOARD: Access/Refresh token fetch failure, "
                    f"enterprise_customer_uuid: {channel_config.enterprise_customer.uuid}, "
                    f"blackboard_base_url: {channel_config.blackboard_base_url}, "
                    f"auth_response_text: {auth_response.text}"
                    f"config_last_modified: {channel_config.modified}",
                    auth_response.status_code,
                )
            try:
                data = auth_response.json()
                # do not forget to save the new refresh token otherwise subsequent requests will fail
                if "refresh_token" not in data:
                    self._log_info("Server did not return refresh_token, keeping existing one")
                else:
                    # refresh token was returned by server, needs to be used
                    fetched_refresh_token = data["refresh_token"]
                    if not fetched_refresh_token.strip():
                        # we are out of luck, can't use this invalid token
                        # and we probably can't get unstuck without customer re-doing oauth url workflow
                        self._log_error("Fetched a new refresh token, but it was empty, not using it!")
                    else:
                        channel_config.refresh_token = fetched_refresh_token
                        self._log_info("Fetched a new refresh token, replacing current one")
                        channel_config.save()
                        # We do not want any fail-prone code in this atomic block after this line
                        # it's because if something else fails, it will roll back the just-saved
                        # refresh token!
                return data['access_token'], data["expires_in"]
            except (KeyError, ValueError) as error:
                raise ClientError(auth_response.text, auth_response.status_code) from error

    def _create_auth_header(self):
        """
        auth header in oauth2 token format as required by blackboard doc
        """
        app_key = self.enterprise_configuration.decrypted_client_id
        if not app_key:
            if not self.global_blackboard_config.app_key:
                raise ClientError(
                    "Failed to generate oauth access token: Client ID required.",
                    HTTPStatus.INTERNAL_SERVER_ERROR.value
                )
            app_key = self.global_blackboard_config.app_key

        app_secret = self.enterprise_configuration.decrypted_client_secret
        if not app_secret:
            if not self.global_blackboard_config.app_secret:
                raise ClientError(
                    "Failed to generate oauth access token: Client secret required.",
                    HTTPStatus.INTERNAL_SERVER_ERROR.value
                )
            app_secret = self.global_blackboard_config.app_secret
        return f"Basic {base64.b64encode(f'{app_key}:{app_secret}'.encode('utf-8')).decode()}"

    def generate_blackboard_course_id(self, external_id):
        """
        A course_id suitable for use with blackboard
        """
        return str(abs(hash(external_id)))

    def generate_blackboard_gradebook_column_data(self, external_id, grade_column_name, points_possible,
                                                  include_in_calculations=False):
        """
        Properly formatted json data to create a new gradebook column in a blackboard course

        Note: Potential customization here per-customer, if the need arises.
        """
        return {
            "externalId": external_id,
            "name": grade_column_name,
            "displayName": grade_column_name,
            "description": GRADEBOOK_COLUMN_DESC,
            "externalGrade": False,
            "score": {
                "possible": points_possible
            },
            "availability": {
                "available": "Yes"
            },
            "grading": {
                "type": "Manual",
                "scoringModel": "Last",
                "anonymousGrading": {
                    "type": "None",
                }
            },
            "includeInCalculations": include_in_calculations,
        }

    def generate_gradebook_url(self, course_id):
        """
        Blackboard API url helper method.
        Path: Get course gradebook
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=GRADEBOOK_PATH.format(course_id=course_id),
        )

    def generate_enrollment_url(self, course_id):
        """
        Blackboard API url helper method.
        Path: Get course enrollments
        """
        # By including `expand=user` we get access to the user's contact info
        return '{base}{path}?expand=user'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=ENROLLMENT_PATH.format(course_id=course_id),
        )

    def generate_course_create_url(self):
        """
        Url to create a course
        """
        return "{base}{path}".format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=COURSES_V3_PATH,
        )

    def generate_course_update_url(self, course_id):
        """
        Url to update one course
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=COURSE_V3_PATH.format(course_id=course_id),
        )

    def generate_courses_url(self):
        """
        Blackboard API url helper method.
        Path: Get course courses
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=COURSE_PATH,
        )

    def generate_create_grade_column_url(self, course_id):
        """
        Blackboard API url helper method.
        Path: Create course grade column
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=POST_GRADE_COLUMN_PATH.format(course_id=course_id),
        )

    def generate_update_grade_column_url(self, course_id, column_id):
        """
        Blackboard API url helper method.
        Path: Update course grade column
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=PATCH_GRADE_COLUMN_PATH.format(
                course_id=course_id,
                column_id=column_id,
            ),
        )

    def generate_post_users_grade_url(self, course_id, column_id, user_id):
        """
        Blackboard API url helper method.
        Path: User's grade column entry
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=POST_GRADE_PATH.format(
                course_id=course_id,
                column_id=column_id,
                user_id=user_id,
            ),
        )

    def generate_create_course_content_url(self, course_id):
        """
        Blackboard API url helper method.
        Path: Course's content objects
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=COURSE_CONTENT_PATH.format(course_id=course_id)
        )

    def generate_create_course_content_child_url(self, course_id, content_id):
        """
        Blackboard API url helper method.
        Path: Course content's children objects
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=COURSE_CONTENT_CHILDREN_PATH.format(course_id=course_id, content_id=content_id)
        )

    def generate_course_content_delete_url(self, course_id, content_id):
        """
        Blackboard API url helper method.
        Path: Course content deletion
        """
        return '{base}{path}'.format(
            base=self.enterprise_configuration.blackboard_base_url,
            path=COURSE_CONTENT_DELETE_PATH.format(course_id=course_id, content_id=content_id)
        )

    def _get(self, url, data=None):
        """
        Returns request's get response and raises Client Errors if appropriate.
        """
        start_time = time.time()
        get_response = self.session.get(url, params=data)
        time_taken = time.time() - start_time
        stringify_and_store_api_record(
            enterprise_customer=self.enterprise_configuration.enterprise_customer,
            enterprise_customer_configuration_id=self.enterprise_configuration.id,
            endpoint=url,
            data=data,
            time_taken=time_taken,
            status_code=get_response.status_code,
            response_body=get_response.text,
            channel_name=self.enterprise_configuration.channel_code()
        )
        if get_response.status_code >= 400:
            raise ClientError(get_response.text, get_response.status_code)
        return get_response

    def _patch(self, url, data):
        """
        Returns request's patch response and raises Client Errors if appropriate.
        """
        start_time = time.time()
        patch_response = self.session.patch(url, json=data)
        time_taken = time.time() - start_time
        stringify_and_store_api_record(
            enterprise_customer=self.enterprise_configuration.enterprise_customer,
            enterprise_customer_configuration_id=self.enterprise_configuration.id,
            endpoint=url,
            data=data,
            time_taken=time_taken,
            status_code=patch_response.status_code,
            response_body=patch_response.text,
            channel_name=self.enterprise_configuration.channel_code()
        )
        if patch_response.status_code >= 400:
            raise ClientError(patch_response.text, patch_response.status_code)
        return patch_response

    def _post(self, url, data):
        """
        Returns request's post response and raises Client Errors if appropriate.
        """
        start_time = time.time()
        post_response = self.session.post(url, json=data)
        time_taken = time.time() - start_time
        stringify_and_store_api_record(
            enterprise_customer=self.enterprise_configuration.enterprise_customer,
            enterprise_customer_configuration_id=self.enterprise_configuration.id,
            endpoint=url,
            data=data,
            time_taken=time_taken,
            status_code=post_response.status_code,
            response_body=post_response.text,
            channel_name=self.enterprise_configuration.channel_code()
        )

        if post_response.status_code >= 400:
            raise ClientError(post_response.text, post_response.status_code)
        return post_response

    def _delete(self, url):
        """
        Returns request's delete response and raises Client Errors if appropriate.
        """
        start_time = time.time()
        response = self.session.delete(url)
        time_taken = time.time() - start_time
        stringify_and_store_api_record(
            enterprise_customer=self.enterprise_configuration.enterprise_customer,
            enterprise_customer_configuration_id=self.enterprise_configuration.id,
            endpoint=url,
            data='',
            time_taken=time_taken,
            status_code=response.status_code,
            response_body=response.text,
            channel_name=self.enterprise_configuration.channel_code()
        )
        if response.status_code >= 400:
            raise ClientError(response.text, response.status_code)
        return response

    def _get_bb_user_id_from_enrollments(self, user_email, course_id):
        """
        Helper method to retrieve a user's Blackboard ID from a list of enrollments in a
        Blackboard class.

        Parameters:
        -----------
            user_email (str): The shared email of the user for both Blackboard and edX
            course_id (str): The Blackboard course ID of which to search enrollments.
        """
        enrollments_response = self._get(self.generate_enrollment_url(course_id)).json()
        for enrollment in enrollments_response.get('results'):
            # No point in checking non-students
            if enrollment.get('courseRoleId') == 'Student':
                contact = enrollment.get('user').get('contact')
                if contact.get('email') == user_email:
                    return enrollment.get('userId')
        raise ClientError(
            'Could not find user={} enrolled in Blackboard course={}'.format(user_email, course_id),
            HTTPStatus.NOT_FOUND.value
        )

    def _get_or_create_integrated_grade_column(self, bb_course_id, grade_column_name, external_id, points_possible=100,
                                               include_in_calculations=False):
        """
        Helper method to search an edX integrated Blackboard course for the designated edX grade column.
        If the column does not yet exist within the course, create it.

        Parameters:
        -----------
            bb_course_id (str): The Blackboard course ID in which to search for the edX final grade,
            grade column.
        """
        gradebook_column_url = self.generate_gradebook_url(bb_course_id)
        grade_column_id = None
        more_pages_present = True
        current_page_count = 0

        # Page count of 250 is a preventative of infinite while loops
        while more_pages_present and current_page_count <= PAGE_TRAVERSAL_LIMIT:
            grade_column_response = self._get(gradebook_column_url)
            parsed_response = grade_column_response.json()
            grade_columns = parsed_response.get('results')

            for grade_column in grade_columns:
                if grade_column.get('externalId') == external_id:
                    grade_column_id = grade_column.get('id')
                    # if includeInCalculations has legacy value, correct it
                    if (
                        grade_column.get('includeInCalculations') and
                        grade_column.get('includeInCalculations') != include_in_calculations
                    ):
                        calculations_data = {"includeInCalculations": include_in_calculations}
                        self._patch(
                            self.generate_update_grade_column_url(bb_course_id, grade_column_id),
                            calculations_data
                        )
                    break
            # Blackboard's pagination is returned within the response json if it exists
            # Example:
            # Response = {
            #     'results': [{
            #         'id': 1,
            #         ...
            #     }, {
            #         'id', 2,
            #         ...
            #     }],
            #     'paging': {
            #         'nextPage': '/learn/api/public/v1/courses/<courseID>/gradebook/columns?offset=200'
            #     }
            # }
            if parsed_response.get('paging') and not grade_column_id:
                gradebook_column_url = '{}{}'.format(
                    self.enterprise_configuration.blackboard_base_url,
                    parsed_response.get('paging').get('nextPage'),
                )
                current_page_count += 1
            else:
                more_pages_present = False

        if current_page_count == PAGE_TRAVERSAL_LIMIT:
            LOGGER.warning(
                generate_formatted_log(
                    self.enterprise_configuration.channel_code(),
                    self.enterprise_configuration.enterprise_customer.uuid,
                    None,
                    None,
                    f'Max page limit hit while traversing blackboard API for course={external_id}'
                )
            )

        if not grade_column_id:
            grade_column_data = self.generate_blackboard_gradebook_column_data(
                external_id, grade_column_name, points_possible, include_in_calculations
            )
            response = self._post(self.generate_create_grade_column_url(bb_course_id), grade_column_data)
            parsed_response = response.json()
            grade_column_id = parsed_response.get('id')

            # Sanity check that we created the grade column properly
            if not grade_column_id:
                raise ClientError(
                    'Something went wrong while create edX integration grade column for course={}.'.format(
                        bb_course_id
                    ),
                    HTTPStatus.INTERNAL_SERVER_ERROR.value
                )

        return grade_column_id

    def _submit_grade_to_blackboard(self, grade, course_id, grade_column_id, user_id):
        """
        Helper method to post learner data to the integrated blackboard course and then validate the submission of
        grade reporting.
        """
        grade_percent = {'score': grade}
        response = self._patch(
            self.generate_post_users_grade_url(course_id, grade_column_id, user_id),
            grade_percent
        )

        if response.json().get('score') != grade:
            raise ClientError(
                'Failed to post new grade for user={} enrolled in course={}'.format(user_id, course_id),
                HTTPStatus.INTERNAL_SERVER_ERROR.value
            )

        return response

    def _resolve_blackboard_course_id(self, external_id):
        """
        Extract course id from blackboard, given it's externalId
        """
        params = {'externalId': external_id}
        courses_responses = self._get(self.generate_courses_url(), params).json()
        course_response = courses_responses.get('results')

        for course in course_response:
            if course.get('externalId') == external_id:
                return course.get('id')
        return None

    def delete_course_content_from_blackboard(self, bb_course_id, bb_content_id):
        """
        Error handling Helper method that will delete any successful content posts that occurred during
        the error'd transmission.
        """
        delete_course_content_url = urljoin(
            self.enterprise_configuration.blackboard_base_url,
            COURSE_CONTENT_DELETE_PATH.format(course_id=bb_course_id, content_id=bb_content_id),
        )
        resp = self._delete(delete_course_content_url)
        return resp

    def delete_course_from_blackboard(self, bb_course_id, bb_content_id=None):
        """
        Error handling Helper method that will delete any already made, successful post requests that occurred during
        the content creation flow.
        """
        # Remove the content if it exists
        if bb_content_id:
            self.delete_course_content_from_blackboard(bb_course_id, bb_content_id)

        # Remove the course
        delete_course_path = urljoin(
            self.enterprise_configuration.blackboard_base_url,
            COURSE_V3_PATH.format(course_id=bb_course_id),
        )
        resp = self._delete(delete_course_path)
        return resp

    def update_integration_content_for_course(self, bb_course_id, channel_metadata_item):
        """
        Helper method to update the blackboard course content page with the integration custom information.
        Differs from course customization creation in that we don't delete the course if things fail.
        """
        content_url = self.generate_create_course_content_url(bb_course_id)
        error_message = ''
        try:
            content_resp = self._post(content_url, channel_metadata_item.get('course_content_metadata'))
            bb_content_id = content_resp.json().get('id')
        except ClientError as error:
            bb_content_id = None
            error_message = ' and received error response={}'.format(error.message)

        if not bb_content_id:
            error_message = '' if not error_message else error_message
            raise ClientError(
                'Something went wrong while creating course content object on Blackboard. Could not retrieve content '
                'ID{}.'.format(error_message),
                HTTPStatus.NOT_FOUND.value
            )

        try:
            content_child_url = self.generate_create_course_content_child_url(bb_course_id, bb_content_id)
            course_fully_created_response = self._post(
                content_child_url, channel_metadata_item.get('course_child_content_metadata')
            )
            content_child_id = course_fully_created_response.json().get('id')
        except ClientError as error:
            content_child_id = None
            error_message = ' and received error response={}'.format(error.message)

        if not content_child_id:
            error_message = '' if not error_message else error_message
            error_handling_response = self.delete_course_content_from_blackboard(bb_course_id, bb_content_id)

            raise ClientError(
                'Something went wrong while creating course content child object on Blackboard. Could not retrieve a '
                'content child ID and got error response={}. Deleted associated course content shell with response: '
                '(status_code={}, body={})'.format(
                    error_message,
                    error_handling_response.status_code,
                    error_handling_response.text
                ),
                HTTPStatus.NOT_FOUND.value
            )

        return course_fully_created_response

    def create_integration_content_for_course(self, bb_course_id, channel_metadata_item):
        """
        Helper method to generate the default blackboard course content page with the integration custom information.
        """
        content_url = self.generate_create_course_content_url(bb_course_id)
        error_message = ''
        try:
            content_resp = self._post(content_url, channel_metadata_item.get('course_content_metadata'))
            bb_content_id = content_resp.json().get('id')
        except ClientError as error:
            bb_content_id = None
            error_message = ' and received error response={}'.format(error.message)

        if not bb_content_id:
            error_handling_response = self.delete_course_from_blackboard(bb_course_id)
            error_message = '' if not error_message else error_message
            raise ClientError(
                'Something went wrong while creating course content object on Blackboard. Could not retrieve content '
                'ID{}. Deleted course with response (status_code={}, body={})'.format(
                    error_message,
                    error_handling_response.status_code,
                    error_handling_response.text
                ),
                HTTPStatus.NOT_FOUND.value
            )

        try:
            content_child_url = self.generate_create_course_content_child_url(bb_course_id, bb_content_id)
            course_fully_created_response = self._post(
                content_child_url, channel_metadata_item.get('course_child_content_metadata')
            )
            content_child_id = course_fully_created_response.json().get('id')
        except ClientError as error:
            content_child_id = None
            error_message = ' and received error response={}'.format(error.message)

        if not content_child_id:
            error_message = '' if not error_message else error_message
            error_handling_response = self.delete_course_from_blackboard(bb_course_id, bb_content_id)
            raise ClientError(
                'Something went wrong while creating course content child object on Blackboard. Could not retrieve a '
                'content child ID and got error response={}. Deleted associated course and content with response: '
                '(status_code={}, body={})'.format(
                    error_message,
                    error_handling_response.status_code,
                    error_handling_response.text
                ),
                HTTPStatus.NOT_FOUND.value
            )

        return course_fully_created_response

    def cleanup_duplicate_assignment_records(self, courses):
        """
        Not implemented yet.
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Blackboard integrated channel does not yet support assignment deduplication."
            )
        )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('enterprise', '0235_enterprisecustomeradmin'),
    ]
    operations = [
        # Custom SQL to add view that unions EnterpriseCustomerUser and PendingEnterpriseCustomerUser records
        # together to facilitate querying both simultaneously
        migrations.RunSQL(
            sql="""
                CREATE VIEW view_enterprise_customer_support_users AS
                    with admins as (
                        SELECT era.id, era.enterprise_customer_id, era.user_id
                        FROM enterprise_systemwideenterpriseuserroleassignment AS era
                        INNER JOIN enterprise_systemwideenterpriserole AS ra ON era.role_id = ra.id
                        WHERE ra.name = 'enterprise_admin'
                    )
                    SELECT ecu.enterprise_customer_id,
                    au.email AS user_email,
                    FALSE AS is_pending,
                    ecu.user_id,
                    au.username,
                    au.first_name,
                    au.last_name,
                    au.is_staff,
                    au.is_active,
                    au.date_joined,
                    CASE WHEN ad.id IS NULL THEN false ELSE true END AS is_admin
                    FROM enterprise_enterprisecustomeruser AS ecu
                    INNER JOIN auth_user AS au ON au.id = ecu.user_fk
                    LEFT JOIN admins AS ad ON ad.enterprise_customer_id = ecu.enterprise_customer_id AND ad.user_id = au.id
                
                    UNION
                
                    SELECT pecu.enterprise_customer_id,
                    pecu.user_email,
                    TRUE AS is_pending,
                    0 AS user_id,
                    '' AS username,
                    '' AS first_name,
                    '' AS last_name,
                    FALSE AS is_staff,
                    FALSE AS is_active,
                    NULL AS date_joined,
                    CASE WHEN pecau.id IS NULL THEN false ELSE true END AS is_admin
                    FROM enterprise_pendingenterprisecustomeruser AS pecu
                    LEFT JOIN enterprise_pendingenterprisecustomeradminuser AS pecau on pecu.user_email = pecau.user_email
            """,
            reverse_sql="""
                DROP VIEW view_enterprise_customer_support_users
            """
        ),
    ]



"""
Waffle toggles for enterprise features within the LMS.
"""

from edx_toggles.toggles import WaffleFlag

ENTERPRISE_NAMESPACE = 'enterprise'
ENTERPRISE_LOG_PREFIX = 'Enterprise: '

# .. toggle_name: enterprise.TOP_DOWN_ASSIGNMENT_REAL_TIME_LCM
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables top-down assignment
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2023-09-15
TOP_DOWN_ASSIGNMENT_REAL_TIME_LCM = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.top_down_assignment_real_time_lcm',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.feature_prequery_search_suggestions
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables prequery search suggestions
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2024-01-31
FEATURE_PREQUERY_SEARCH_SUGGESTIONS = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.feature_prequery_search_suggestions',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)


# .. toggle_name: enterprise.enterprise_customer_support_tool
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables the enterprise customer support tool
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2024-07-17
ENTERPRISE_CUSTOMER_SUPPORT_TOOL = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.enterprise_customer_support_tool',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.enterprise_learner_bff_enabled
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables the enterprise learner BFF
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2024-12-16
ENTERPRISE_LEARNER_BFF_ENABLED = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.learner_bff_enabled',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.admin_portal_learner_profile_view_enabled
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables an admin to view a learner's profile
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2025-02-12
ADMIN_PORTAL_LEARNER_PROFILE_VIEW_ENABLED = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.admin_portal_learner_profile_view_enabled',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.catalog_query_search_filters_enabled
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables filtering search results by catalog queries vs. enterprise-specific attributes.
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2025-03-02
CATALOG_QUERY_SEARCH_FILTERS_ENABLED = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.catalog_query_search_filters_enabled',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.enterprise_admin_onboarding
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables the admin onboarding tour on the admin-portal.
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2025-05-12
ENTERPRISE_ADMIN_ONBOARDING = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.enterprise_admin_onboarding',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.edit_highlights_enabled
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables the Edit Highlights experience.
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2025-07-21
EDIT_HIGHLIGHTS_ENABLED = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.edit_highlights_enabled',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)

# .. toggle_name: enterprise.invite_admins_enabled
# .. toggle_implementation: WaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables the invite admins on people management page.
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2026-01-06
INVITE_ADMINS_ENABLED = WaffleFlag(
    f'{ENTERPRISE_NAMESPACE}.invite_admins_enabled',
    __name__,
    ENTERPRISE_LOG_PREFIX,
)


def top_down_assignment_real_time_lcm():
    """
    Returns whether top-down assignment and real time LCM feature flag is enabled.
    """
    return TOP_DOWN_ASSIGNMENT_REAL_TIME_LCM.is_enabled()


def feature_prequery_search_suggestions():
    """
    Returns whether the prequery search suggestion feature flag is enabled.
    """
    return FEATURE_PREQUERY_SEARCH_SUGGESTIONS.is_enabled()


def enterprise_customer_support_tool():
    """
    Returns whether the enterprise customer support tool is enabled.
    """
    return ENTERPRISE_CUSTOMER_SUPPORT_TOOL.is_enabled()


def enterprise_learner_bff_enabled():
    """
    Returns whether the enterprise learner BFF is enabled.
    """
    return ENTERPRISE_LEARNER_BFF_ENABLED.is_enabled()


def admin_portal_learner_profile_view_enabled():
    """
    Returns whether the learner profile view in admin portal is enabled.
    """
    return ADMIN_PORTAL_LEARNER_PROFILE_VIEW_ENABLED.is_enabled()


def catalog_query_search_filters_enabled():
    """
    Returns whether the catalog query search filters feature flag is enabled.
    """
    return CATALOG_QUERY_SEARCH_FILTERS_ENABLED.is_enabled()


def enterprise_admin_onboarding_enabled():
    """
    Returns whether the admin onboarding feature flag is enabled.
    """
    return ENTERPRISE_ADMIN_ONBOARDING.is_enabled()


def enterprise_edit_highlights_enabled():
    """
    Returns whether the edit highlights feature flag is enabled.
    """
    return EDIT_HIGHLIGHTS_ENABLED.is_enabled()


def enterprise_invite_admins_enabled():
    """
    Returns whether the invite admins feature flag is enabled.
    """
    return INVITE_ADMINS_ENABLED.is_enabled()


def enterprise_features():
    """
    Returns a dict of enterprise Waffle-based feature flags.
    """
    return {
        'top_down_assignment_real_time_lcm': top_down_assignment_real_time_lcm(),
        'feature_prequery_search_suggestions': feature_prequery_search_suggestions(),
        'enterprise_customer_support_tool': enterprise_customer_support_tool(),
        'enterprise_learner_bff_enabled': enterprise_learner_bff_enabled(),
        'admin_portal_learner_profile_view_enabled': admin_portal_learner_profile_view_enabled(),
        'catalog_query_search_filters_enabled': catalog_query_search_filters_enabled(),
        'enterprise_admin_onboarding_enabled': enterprise_admin_onboarding_enabled(),
        'enterprise_edit_highlights_enabled': enterprise_edit_highlights_enabled(),
        'enterprise_invite_admins_enabled': enterprise_invite_admins_enabled(),
    }



"""
Views for the `enterprise-sso-user-info` API endpoint.
"""

from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from django.core.exceptions import ValidationError

from enterprise.api.v1.serializers import EnterpriseSSOUserInfoRequestSerializer
from enterprise.logging import getEnterpriseLogger
from enterprise.models import EnterpriseCustomer
from enterprise.utils import get_user_details

LOGGER = getEnterpriseLogger(__name__)


class EnterpriseSSOUserViewSet(viewsets.ViewSet):
    """
    API views for Enterprise SSO user operations.

    Provides endpoints for retrieving Enterprise SSO user details.
    """
    permission_classes = (IsAuthenticated,)

    @action(detail=False, methods=['get'])
    def user_details(self, request):
        """
        Retrieve Enterprise SSO user details.

        Query Parameters:
            * org_id: ID of the organization to retrieve Enterprise Customer
            * external_user_id: ID of the user to get details for

        Returns:
            Response with Enterprise SSO user details
        """
        serializer = EnterpriseSSOUserInfoRequestSerializer(data=request.query_params)

        if not serializer.is_valid():
            return Response(
                serializer.errors,
                status=status.HTTP_400_BAD_REQUEST
            )

        org_id = serializer.validated_data['org_id']
        external_user_id = serializer.validated_data['external_user_id']

        try:
            # Get Enterprise Customer by org_id
            enterprise_customer = EnterpriseCustomer.active_customers.get(auth_org_id=org_id)

            user_details = get_user_details(enterprise_customer, external_user_id)

            if not user_details:
                return Response(
                    {'error': 'Enterprise SSO user details not found'},
                    status=status.HTTP_404_NOT_FOUND
                )

            return Response(user_details, status=status.HTTP_200_OK)

        except EnterpriseCustomer.DoesNotExist:
            return Response(
                {'error': f'Enterprise customer not found for org_id: {org_id}'},
                status=status.HTTP_404_NOT_FOUND
            )
        except ValidationError as exc:
            return Response(
                {'error': str(exc)},
                status=status.HTTP_400_BAD_REQUEST
            )



"""edX Enterprise's Consent application.

This application provides a generic Consent API that lets the
user get, provide, and revoke consent to an enterprise customer
at some gate.
"""

from django.apps import AppConfig


class ConsentConfig(AppConfig):
    """Configuration for edX Enterprise's Consent application."""

    name = 'consent'
    verbose_name = "Enterprise Consent"

    def ready(self):
        """
        Perform one-time initialization: connect signal handlers.
        """
        import consent.signals  # noqa: F401  pylint: disable=import-outside-toplevel,unused-import



# Generated migration for MariaDB UUID field conversion (Django 5.2)
"""
Migration to convert UUIDField from char(32) to uuid type for MariaDB compatibility.

See: https://www.albertyw.com/note/django-5-mariadb-uuidfield
"""

from django.db import migrations


def apply_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection

    if connection.vendor != 'mysql':
        return

    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return

    with connection.cursor() as cursor:
        cursor.execute("ALTER TABLE blackboard_channel_blackboardenterprisecustomerconfiguration MODIFY uuid uuid NOT NULL")


def reverse_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection

    if connection.vendor != 'mysql':
        return

    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return

    with connection.cursor() as cursor:
        cursor.execute("ALTER TABLE blackboard_channel_blackboardenterprisecustomerconfiguration MODIFY uuid char(32) NOT NULL")


class Migration(migrations.Migration):

    dependencies = [
        ('blackboard', '0024_rename_blackboardlearnerdatatransmissionaudit_enterprise_customer_uuid_plugin_configuration_id_black'),
    ]

    operations = [
        migrations.RunPython(
            code=apply_mariadb_migration,
            reverse_code=reverse_mariadb_migration,
        ),
    ]



"""
Viewsets for integrated_channels/v1/cornerstone/
"""
from logging import getLogger

from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import permissions, viewsets
from rest_framework.authentication import SessionAuthentication
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_404_NOT_FOUND
from rest_framework.views import APIView

from django.contrib import auth
from django.db import transaction

from enterprise.api.throttles import ServiceUserThrottle
from enterprise.utils import get_enterprise_customer_or_404, get_enterprise_customer_user, localized_utcnow
from integrated_channels.api.v1.mixins import PermissionRequiredForIntegratedChannelMixin
from integrated_channels.cornerstone.models import CornerstoneEnterpriseCustomerConfiguration
from integrated_channels.cornerstone.utils import create_cornerstone_learner_data

from .serializers import CornerstoneConfigSerializer

LOGGER = getLogger(__name__)
User = auth.get_user_model()


class CornerstoneConfigurationViewSet(PermissionRequiredForIntegratedChannelMixin, viewsets.ModelViewSet):
    """Viewset for CornerstoneEnterpriseCustomerConfiguration"""
    serializer_class = CornerstoneConfigSerializer
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'enterprise.can_access_admin_dashboard'

    configuration_model = CornerstoneEnterpriseCustomerConfiguration


class CornerstoneLearnerInformationView(APIView):
    """Viewset for saving information of a cornerstone learner"""
    permission_classes = (permissions.IsAuthenticated,)
    authentication_classes = (JwtAuthentication, SessionAuthentication,)
    throttle_classes = (ServiceUserThrottle,)

    def post(self, request):
        """
            An endpoint to save a cornerstone learner information received from frontend.
            integrated_channels/api/v1/cornerstone/save-learner-information
            Requires a JSON object in the following format:
                {
                    "courseKey": "edX+DemoX",
                    "enterpriseUUID": "enterprise-uuid-goes-right-here",
                    "userGuid": "user-guid-from-csod",
                    "callbackUrl": "https://example.com/csod/callback/1",
                    "sessionToken": "123123123",
                    "subdomain": "edx.csod.com"
                }
        """
        user_id = request.user.id
        enterprise_customer_uuid = request.data.get('enterpriseUUID')
        enterprise_customer = get_enterprise_customer_or_404(enterprise_customer_uuid)
        course_key = request.data.get('courseKey')
        with transaction.atomic():
            csod_user_guid = request.data.get('userGuid')
            csod_callback_url = request.data.get('callbackUrl')
            csod_session_token = request.data.get('sessionToken')
            csod_subdomain = request.data.get("subdomain")

            if csod_session_token and csod_subdomain:
                LOGGER.info(
                    f'integrated_channel=CSOD, '
                    f'integrated_channel_enterprise_customer_uuid={enterprise_customer_uuid}, '
                    f'integrated_channel_lms_user={user_id}, '
                    f'integrated_channel_course_key={course_key}, '
                    'saving CSOD learner information'
                )
                cornerstone_customer_configuration = \
                    CornerstoneEnterpriseCustomerConfiguration.get_by_customer_and_subdomain(
                        enterprise_customer=enterprise_customer,
                        customer_subdomain=csod_subdomain
                    )
                if cornerstone_customer_configuration:
                    # check if request user is linked as a learner with the given enterprise before savin anything
                    enterprise_customer_user = get_enterprise_customer_user(user_id, enterprise_customer_uuid)
                    if enterprise_customer_user:
                        # saving session token in enterprise config to access cornerstone apis
                        cornerstone_customer_configuration.session_token = csod_session_token
                        cornerstone_customer_configuration.session_token_modified = localized_utcnow()
                        cornerstone_customer_configuration.save()
                        # saving learner information received from cornerstone
                        create_cornerstone_learner_data(
                            user_id,
                            csod_user_guid,
                            csod_session_token,
                            csod_callback_url,
                            csod_subdomain,
                            cornerstone_customer_configuration,
                            course_key
                        )
                    else:
                        LOGGER.error(
                            f'integrated_channel=CSOD, '
                            f'integrated_channel_enterprise_customer_uuid={enterprise_customer_uuid}, '
                            f'integrated_channel_lms_user={user_id}, '
                            f'integrated_channel_course_key={course_key}, '
                            f'user is not linked to the given enterprise'
                        )
                        message = (f'Cornerstone information could not be saved for learner with user_id={user_id}'
                                   f'because user is not linked to the given enterprise {enterprise_customer_uuid}')
                        return Response(data={'error': message}, status=HTTP_404_NOT_FOUND)
                else:
                    LOGGER.error(
                        f'integrated_channel=CSOD, '
                        f'integrated_channel_enterprise_customer_uuid={enterprise_customer_uuid}, '
                        f'integrated_channel_lms_user={user_id}, '
                        f'integrated_channel_course_key={course_key}, '
                        f'unable to find cornerstone config matching subdomain {csod_subdomain}'
                    )
                    message = (f'Cornerstone information could not be saved for learner with user_id={user_id}'
                               f'because no config exist with the subdomain {csod_subdomain}')
                    return Response(data={'error': message}, status=HTTP_404_NOT_FOUND)
            return Response(status=HTTP_200_OK)



# Generated by Django 3.2.22 on 2024-01-25 09:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0031_cornerstoneapirequestlogs'),
    ]

    operations = [
        migrations.DeleteModel(
            name='CornerstoneAPIRequestLogs',
        ),
    ]



"""
Management command for assigning enterprise_learner roles
to existing linked enterprise users that are missing them.
"""

import logging
from time import sleep

from django.contrib import auth
from django.core.management.base import BaseCommand
from django.db import DatabaseError, models, transaction

from enterprise.models import EnterpriseCustomerUser
from integrated_channels.utils import batch_by_pk

log = logging.getLogger(__name__)
User = auth.get_user_model()


def safe_bulk_update(entries, properties, max_retries, manager):
    """Performs bulk_update with retry logic."""
    for attempt in range(1, max_retries + 1):
        try:
            with transaction.atomic():
                manager.bulk_update(
                    entries, properties
                )
            return len(entries)
        except DatabaseError as e:
            wait_time = 2 ** attempt
            log.warning(f"Attempt {attempt}/{max_retries} failed: {e}. Retrying in {wait_time}s.")
            sleep(wait_time)

    raise Exception(f"Bulk update failed after {max_retries} retries.")


def _fetch_and_update_in_batches(manager, batch_limit, batch_sleep, max_retries, ModelClass):
    """
    Fetches and updates records in batches.
    Only loads and updates a subset of records at a time to avoid memory and performance issues.
    Note: you cannot use django's queryset.iterator() method
    as MySQL does not support it and will still load everything into memory.
    """
    batch_counter = 1
    total_processed = 0

    for batch in batch_by_pk(
        ModelClass,
        extra_filter=models.Q(user_fk__isnull=True),
        batch_size=batch_limit,
        model_manager=manager,
    ):
        log.info(f"Processing batch {batch_counter}...")
        for ecu in batch:
            if isinstance(ModelClass._meta.get_field('user_fk'), models.ForeignKey):
                ecu.user_fk_id = ecu.user_id
            else:
                ecu.user_fk = ecu.user_id
        safe_bulk_update(batch, ['user_fk'], max_retries, manager=manager)
        total_processed += len(batch)
        log.info(f'Processed {total_processed} records.')
        sleep(batch_sleep)
        batch_counter += 1

    log.info(f"Final batch processed. Total {total_processed} records updated.")
    return True


class Command(BaseCommand):
    """
    Management command to copy `user_id` values to the foreign key `user_fk` field
    in the enterprise_customer_user table.

    Example usage:
        $ ./manage.py backfill_ecu_table_user_foreign_key
    """
    help = '''
    Goes through the Enterprise Customer User table in batches
    and copies the user_id to the user_fk foreign key.
    '''

    def add_arguments(self, parser):
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            '--batch-limit',
            action='store',
            dest='batch_limit',
            default=250,
            help='Number of records in each batch.',
            type=int,
        )
        parser.add_argument(
            '--max-retries',
            action='store',
            dest='max_retries',
            default=5,
            help='Max retries for each batch.',
            type=int,
        )
        parser.add_argument(
            '--batch-sleep',
            action='store',
            dest='batch_sleep',
            default=0.1,
            help='How long to sleep between batches.',
            type=float,
        )

    def backfill_ecu_table_user_foreign_key(self, options):
        """
        Backfills user_fk from user_id in batches with timeout, retries, and progress reporting.
        """
        batch_limit = options.get('batch_limit', 250)
        max_retries = options.get('max_retries', 5)
        batch_sleep = options.get('batch_sleep', 0.1)

        _fetch_and_update_in_batches(
            manager=EnterpriseCustomerUser.all_objects,
            batch_limit=batch_limit,
            batch_sleep=batch_sleep,
            max_retries=max_retries,
            ModelClass=EnterpriseCustomerUser,
        )

        _fetch_and_update_in_batches(
            manager=EnterpriseCustomerUser.history,
            batch_limit=batch_limit,
            batch_sleep=batch_sleep,
            max_retries=max_retries,
            ModelClass=EnterpriseCustomerUser,
        )

    def handle(self, *_args, **options):
        """
        Entry point for management command execution.
        """
        log.info('Starting backfilling ECU user_fk field from user_id!')

        self.backfill_ecu_table_user_foreign_key(options)

        log.info('Successfully finished backfilling ECU user_fk field from user_id!')



# Generated by Django 2.2.18 on 2021-10-11 17:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0143_auto_20210908_0559'),
    ]

    operations = [
        migrations.AlterField(
            model_name='adminnotification',
            name='text',
            field=models.CharField(help_text='Notification banner text which will appear to enterprise admin on admin portal. You can enter maximum of 512 characters. This text support markdown. See https://commonmark.org/help/ for the supported markdown features.', max_length=512),
        ),
    ]



# Generated by Django 3.2.15 on 2022-09-29 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0163_auto_20220928_1611'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecatalogquery',
            name='include_exec_ed_2u_courses',
            field=models.BooleanField(default=False, help_text="Specifies whether the catalog is allowed to include exec ed (2U) courses.  This means that, when the content_filter specifies that 'course' content types should be included in the catalog, executive-education-2u course types won't be excluded from the content of the associated catalog."),
        ),
    ]



# Generated by Django 3.2.19 on 2023-05-17 15:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0171_auto_20230503_1413'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecourseenrollment',
            name='unenrolled',
            field=models.BooleanField(blank=True, db_index=True, default=None, help_text="Specifies whether the enterprise course enrollment's course enrollment object was unenrolled.", null=True),
        ),
        migrations.AddField(
            model_name='historicalenterprisecourseenrollment',
            name='unenrolled',
            field=models.BooleanField(blank=True, db_index=True, default=None, help_text="Specifies whether the enterprise course enrollment's course enrollment object was unenrolled.", null=True),
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-28 14:58

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0153_add_enable_browse_and_request'),
        ('integrated_channel', '0009_auto_20220325_1757'),
    ]

    operations = [
        migrations.CreateModel(
            name='GenericEnterpriseCustomerPluginConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('display_name', models.CharField(blank=True, default='', help_text='A configuration nickname.', max_length=255)),
                ('idp_id', models.CharField(blank=True, default='', help_text='If provided, will be used as IDP slug to locate remote id for learners', max_length=255)),
                ('active', models.BooleanField(help_text='Is this configuration active?')),
                ('transmission_chunk_size', models.IntegerField(default=500, help_text='The maximum number of data items to transmit to the integrated channel with each request.')),
                ('channel_worker_username', models.CharField(blank=True, default='', help_text='Enterprise channel worker username to get JWT tokens for authenticating LMS APIs.', max_length=255)),
                ('catalogs_to_transmit', models.TextField(blank=True, default='', help_text='A comma-separated list of catalog UUIDs to transmit. If blank, all customer catalogs will be transmitted. If there are overlapping courses in the customer catalogs, the overlapping course metadata will be selected from the newest catalog.')),
                ('disable_learner_data_transmissions', models.BooleanField(default=False, help_text='When set to True, the configured customer will no longer receive learner data transmissions, both scheduled and signal based', verbose_name='Disable Learner Data Transmission')),
                ('enterprise_customer', models.ForeignKey(help_text='Enterprise Customer associated with the configuration.', on_delete=django.db.models.deletion.CASCADE, to='enterprise.enterprisecustomer')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 4.2.15 on 2024-08-26 13:57

from django.db import migrations


def populate_decrypted_client_id_and_secret(apps, schema_editor):  # pragma: no cover
    """
    Populate the decrypted_client_id and decrypted_client_secret fields for all
    existing CanvasEnterpriseCustomerConfiguration
    """
    CanvasEnterpriseCustomerConfiguration = apps.get_model('canvas', 'CanvasEnterpriseCustomerConfiguration')
    for config in CanvasEnterpriseCustomerConfiguration.objects.all():
        config.decrypted_client_id = config.client_id
        config.decrypted_client_secret = config.client_secret
        config.save()


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0036_canvasenterprisecustomerconfiguration_decrypted_client_id_and_more'),
    ]

    operations = [
        migrations.RunPython(populate_decrypted_client_id_and_secret, reverse_code=migrations.RunPython.noop),
    ]



"""
Tests for the `edx-enterprise` utility functions.
"""

import datetime
import unittest
from collections import namedtuple
from unittest import mock

import ddt
import pytz
from faker import Factory as FakerFactory
from pytest import mark, raises

from django.core import mail
from django.core.exceptions import ValidationError
from django.forms.models import model_to_dict
from django.test import override_settings

from enterprise import utils, validators
from enterprise.models import (
    EnterpriseCustomer,
    EnterpriseCustomerBrandingConfiguration,
    EnterpriseCustomerIdentityProvider,
    EnterpriseCustomerUser,
    PendingEnterpriseCustomerUser,
)
from enterprise.utils import (
    ADMIN_ENROLL_EMAIL_TEMPLATE_TYPE,
    SELF_ENROLL_EMAIL_TEMPLATE_TYPE,
    find_enroll_email_template,
)
from integrated_channels.sap_success_factors.models import SAPSuccessFactorsEnterpriseCustomerConfiguration
from test_utils import TEST_UUID, create_items, fake_catalog_api
from test_utils.factories import (
    FAKER,
    EnrollmentNotificationEmailTemplateFactory,
    EnterpriseCourseEnrollmentFactory,
    EnterpriseCustomerFactory,
    EnterpriseCustomerIdentityProviderFactory,
    EnterpriseCustomerUserFactory,
    PendingEnterpriseCustomerUserFactory,
    SiteFactory,
    UserFactory,
)

DATETIME_NOW = datetime.datetime.utcnow()
TEST_PGP_KEY = """
-----BEGIN PGP PRIVATE KEY BLOCK-----
Comment: Alice's OpenPGP Transferable Secret Key
Comment: https://www.ietf.org/id/draft-bre-openpgp-samples-01.html
lFgEXEcE6RYJKwYBBAHaRw8BAQdArjWwk3FAqyiFbFBKT4TzXcVBqPTB3gmzlC/U
b7O1u10AAP9XBeW6lzGOLx7zHH9AsUDUTb2pggYGMzd0P3ulJ2AfvQ4RtCZBbGlj
ZSBMb3ZlbGFjZSA8YWxpY2VAb3BlbnBncC5leGFtcGxlPoiQBBMWCAA4AhsDBQsJ
CAcCBhUKCQgLAgQWAgMBAh4BAheAFiEE64W7X6M6deFelE5j8jFVDE9H444FAl2l
nzoACgkQ8jFVDE9H447pKwD6A5xwUqIDprBzrHfahrImaYEZzncqb25vkLV2arYf
a78A/R3AwtLQvjxwLDuzk4dUtUwvUYibL2sAHwj2kGaHnfICnF0EXEcE6RIKKwYB
BAGXVQEFAQEHQEL/BiGtq0k84Km1wqQw2DIikVYrQrMttN8d7BPfnr4iAwEIBwAA
/3/xFPG6U17rhTuq+07gmEvaFYKfxRB6sgAYiW6TMTpQEK6IeAQYFggAIBYhBOuF
u1+jOnXhXpROY/IxVQxPR+OOBQJcRwTpAhsMAAoJEPIxVQxPR+OOWdABAMUdSzpM
hzGs1O0RkWNQWbUzQ8nUOeD9wNbjE3zR+yfRAQDbYqvtWQKN4AQLTxVJN5X5AWyb
Pnn+We1aTBhaGa86AQ==
=n8OM
-----END PGP PRIVATE KEY BLOCK-----
"""


def mock_get_available_idps(idps):
    """
    Mock method for get_available_idps.
    """
    def _():
        """
        mock function for get_available_idps.
        """
        idp_list = []
        for idp in idps:
            mock_idp = mock.Mock()
            mock_idp.configure_mock(provider_id=idp, name=idp)
            idp_list.append(mock_idp)
        return idp_list
    return _


@mark.django_db
@ddt.ddt
@mark.django_db
class TestEnterpriseUtils(unittest.TestCase):
    """
    Tests for enterprise utility functions.
    """

    def setUp(self):
        """
        Set up test environment.
        """
        super().setUp()
        faker = FakerFactory.create()
        self.provider_id = faker.slug()  # pylint: disable=no-member
        self.uuid = faker.uuid4()  # pylint: disable=no-member
        self.customer = EnterpriseCustomerFactory(uuid=self.uuid)
        self.email_template = EnterpriseCustomerIdentityProviderFactory(
            provider_id=self.provider_id,
            enterprise_customer=self.customer
        )

    @ddt.unpack
    @ddt.data(
        (
            EnterpriseCustomer,
            [
                "enterprise_customer_users",
                "pendingenterprisecustomeruser",
                "branding_configuration",
                "enterprise_customer_identity_providers",
                "enterprise_customer_catalogs",
                "enterprise_enrollment_template",
                "reporting_configurations",
                "pendingenterprisecustomeradminuser",
                "enterprise_customer_consent",
                "data_sharing_consent_page",
                "genericenterprisecustomerpluginconfiguration",
                "contentmetadataitemtransmission",
                "cornerstoneenterprisecustomerconfiguration",
                "degreed2enterprisecustomerconfiguration",
                "degreedenterprisecustomerconfiguration",
                "canvasenterprisecustomerconfiguration",
                "blackboardenterprisecustomerconfiguration",
                "sapsuccessfactorsenterprisecustomerconfiguration",
                "moodleenterprisecustomerconfiguration",
                "xapilrsconfiguration",
                "created",
                "modified",
                "uuid",
                "name",
                "slug",
                "auth_org_id",
                "active",
                "country",
                "integratedchannelapirequestlogs",
                "invite_keys",
                "hide_course_original_price",
                "site",
                "disable_expiry_messaging_for_learner_credit",
                "enable_data_sharing_consent",
                "enforce_data_sharing_consent",
                "enable_audit_enrollment",
                "enable_audit_data_reporting",
                "replace_sensitive_sso_username",
                "enable_autocohorting",
                "customer_type",
                "enable_portal_code_management_screen",
                "enable_portal_reporting_config_screen",
                "enable_portal_subscription_management_screen",
                "enable_portal_saml_configuration_screen",
                "enable_portal_lms_configurations_screen",
                "enable_learner_credit_message_box",
                "enable_universal_link",
                "enable_browse_and_request",
                "enable_learner_portal",
                "enable_learner_portal_offers",
                "enable_portal_learner_credit_management_screen",
                "enable_executive_education_2U_fulfillment",
                "enable_integrated_customer_learner_portal_search",
                "enable_learner_portal_sidebar_message",
                "enable_analytics_screen",
                "enable_slug_login",
                "contact_email",
                "default_contract_discount",
                "default_language",
                "sender_alias",
                "system_wide_role_assignments",
                "reply_to",
                "hide_labor_market_data",
                "chat_gpt_prompts",
                "enable_generation_of_api_credentials",
                "learner_portal_sidebar_content",
                "sso_orchestration_records",
                "enable_pathways",
                "enable_programs",
                "enable_demo_data_for_analytics_and_lpr",
                "enable_academies",
                "enable_one_academy",
                "show_videos_in_learner_portal_search_results",
                "groups",
                "default_enrollment_intentions",
            ]
        ),
        (
            EnterpriseCustomerUser,
            [
                'enterprise_entitlements',
                'adminnotificationread',
                'admin_record',
                "enterprise_enrollments",
                "id",
                "created",
                "memberships",
                "modified",
                "enterprise_customer",
                "user_id",
                "active",
                "linked",
                "is_relinkable",
                "invite_key",
                "should_inactivate_other_customers",
                "user_fk",
            ]
        ),
        (
            EnterpriseCustomerBrandingConfiguration,
            [
                "id",
                "created",
                "modified",
                "enterprise_customer",
                "logo",
                "primary_color",
                "secondary_color",
                "tertiary_color",
            ]
        ),
        (
            EnterpriseCustomerIdentityProvider,
            [
                "id",
                "created",
                "modified",
                "enterprise_customer",
                "provider_id",
                "default_provider",
            ]
        ),
    )
    def test_get_all_field_names(self, model, expected_fields):
        actual_field_names = utils.get_all_field_names(model, excluded=['catalog'])
        assert sorted(actual_field_names) == sorted(expected_fields)

    @ddt.data(
        ("", ""),
        (
            "localhost:18387/api/v1/",
            "localhost:18387/admin/catalogs/catalog/{catalog_id}/change/",
        ),
        (
            "http://localhost:18387/api/v1/",
            "http://localhost:18387/admin/catalogs/catalog/{catalog_id}/change/",
        ),
        (
            "https://prod-site-discovery.example.com/api/v1/",
            "https://prod-site-discovery.example.com/admin/catalogs/catalog/{catalog_id}/change/",
        ),
        (
            "https://discovery-site.subdomain.example.com/api/v1/",
            "https://discovery-site.subdomain.example.com/admin/catalogs/catalog/{catalog_id}/change/",
        ),
    )
    @ddt.unpack
    def test_catalog_admin_url_template(self, catalog_api_url, expected_url):
        """
        Validate that `get_catalog_admin_url_template` utility functions
        returns catalog admin page url template.

        Arguments:
            catalog_api_url (str): course catalog api url coming from DDT data decorator.
            expected_url (str): django admin catalog details page url coming from DDT data decorator.
        """
        with override_settings(COURSE_CATALOG_API_URL=catalog_api_url):
            url = utils.get_catalog_admin_url_template()
            assert url == expected_url

    @ddt.data(
        (7, "", ""),
        (
            7,
            "localhost:18387/api/v1/",
            "localhost:18387/admin/catalogs/catalog/7/change/",
        ),
        (
            7,
            "http://localhost:18387/api/v1/",
            "http://localhost:18387/admin/catalogs/catalog/7/change/",
        ),
        (
            7,
            "https://prod-site-discovery.example.com/api/v1/",
            "https://prod-site-discovery.example.com/admin/catalogs/catalog/7/change/",
        ),
        (
            7,
            "https://discovery-site.subdomain.example.com/api/v1/",
            "https://discovery-site.subdomain.example.com/admin/catalogs/catalog/7/change/",
        ),
    )
    @ddt.unpack
    def test_catalog_admin_url(self, catalog_id, catalog_api_url, expected_url):
        """
        Validate that `get_catalog_admin_url` utility functions returns catalog admin page url.

        Arguments:
            catalog_id (int): catalog id coming from DDT data decorator.
            catalog_api_url (str): course catalog api url coming from DDT data decorator.
            expected_url (str): django admin catalog details page url coming from DDT data decorator.
        """
        with override_settings(COURSE_CATALOG_API_URL=catalog_api_url):
            url = utils.get_catalog_admin_url(catalog_id)
            assert url == expected_url

    @ddt.data(
        (EnterpriseCustomerFactory, [{'uuid': TEST_UUID}], True),
        (None, [{}], False)
    )
    @ddt.unpack
    def test_get_enterprise_customer(self, factory, items, returns_obj):
        if factory:
            create_items(factory, items)
        enterprise_customer = utils.get_enterprise_customer(TEST_UUID)
        if returns_obj:
            self.assertIsNotNone(enterprise_customer)
        else:
            self.assertIsNone(enterprise_customer)

    def test_get_enterprise_customer_user(self):
        user = UserFactory()
        enterprise_customer = EnterpriseCustomerFactory()

        assert utils.get_enterprise_customer_user(user.id, enterprise_customer.uuid) is None

        enterprise_customer_user = EnterpriseCustomerUserFactory(
            user_id=user.id,
            enterprise_customer=enterprise_customer
        )
        assert utils.get_enterprise_customer_user(user.id, enterprise_customer.uuid) == enterprise_customer_user

    def test_find_enroll_email_template_only_fallback_found(self):
        """
        Test the find_enroll_email_template util picks up correct fallback templates
        """
        EnrollmentNotificationEmailTemplateFactory(
            enterprise_customer=None,
            template_type=SELF_ENROLL_EMAIL_TEMPLATE_TYPE
        )
        EnrollmentNotificationEmailTemplateFactory(
            enterprise_customer=None,
            template_type=ADMIN_ENROLL_EMAIL_TEMPLATE_TYPE
        )

        self_fallback = find_enroll_email_template(
            self.customer,
            SELF_ENROLL_EMAIL_TEMPLATE_TYPE
        )
        assert self_fallback is not None
        assert self_fallback.template_type == SELF_ENROLL_EMAIL_TEMPLATE_TYPE

        admin_fallback = find_enroll_email_template(
            self.customer,
            ADMIN_ENROLL_EMAIL_TEMPLATE_TYPE
        )
        assert admin_fallback is not None
        assert admin_fallback.template_type == ADMIN_ENROLL_EMAIL_TEMPLATE_TYPE

    @ddt.data(
        (
            {'class': PendingEnterpriseCustomerUserFactory, 'user_email': 'john@smith.com'},
            {
                'name': 'Demo Course',
                'url': 'http://lms.example.com/courses',
                'type': 'course',
                'start': '2017-01-01'
            },
            'EdX',
            {
                'subject': 'You\'ve been enrolled in Demo Course!',
                'from_email': 'course_staff@example.com',
                'to': [
                    'john@smith.com'
                ],
                'body': (
                    'http://lms.example.com/courses, Demo Course, EdX'
                ),
                'alternatives': [
                    (
                        (
                            '<html><body>http://lms.example.com/courses, Demo Course, EdX</body></html>'
                        ),
                        'text/html'
                    )
                ],
                'attachments': [],
            }
        ),
        (
            {'class': UserFactory, 'first_name': 'John', 'username': '', 'email': 'john@smith.com'},
            {
                'name': 'Enterprise Learning',
                'url': 'http://lms.example.com/courses',
                'type': 'course',
                'start': '2017-06-23'
            },
            'Widgets, Inc',
            {
                'subject': 'You\'ve been enrolled in Enterprise Learning!',
                'from_email': 'course_staff@example.com',
                'to': [
                    'john@smith.com'
                ],
                'body': (
                    'Dear John http://lms.example.com/courses, Enterprise Learning, Widgets, Inc'
                ),
                'alternatives': [
                    (
                        (
                            '<html><body>Dear John http://lms.example.com/courses, '
                            'Enterprise Learning, Widgets, Inc</body></html>'
                        ),
                        'text/html'
                    )
                ],
                'attachments': [],
            }
        ),
        (
            {'class': UserFactory, 'username': 'johnny_boy', 'email': 'john@smith.com', 'first_name': ''},
            {
                "name": "Master of Awesomeness",
                "url": "http://lms.example.com/courses",
                "type": "program",
                "branding": "MicroMaster",
                "start": "2017-04-15",
            },
            'MIT',
            {
                'subject': 'You\'ve been enrolled in Master of Awesomeness!',
                'from_email': 'course_staff@example.com',
                'to': [
                    'john@smith.com'
                ],
                'body': (
                    'Dear johnny_boy http://lms.example.com/courses, Master of Awesomeness, MIT'
                ),
                'alternatives': [
                    (
                        (
                            '<html><body>Dear johnny_boy http://lms.example.com/courses, '
                            'Master of Awesomeness, MIT</body></html>'
                        ),
                        'text/html'
                    )
                ],
                'attachments': [],
            }
        ),
        (
            None,
            {
                'name': 'coursename',
                'url': 'lms.example.com/courses',
                'type': 'program',
                'branding': 'MicroMaster',
                'start': '2017-01-01',
            },
            'EdX',
            {},
        ),
    )
    @ddt.unpack
    def test_send_email_notification_message(
            self,
            user,
            enrolled_in,
            enterprise_customer_name,
            expected_fields,
    ):
        """
        Test that we can successfully render and send an email message.
        """
        enrolled_in['start'] = datetime.datetime.strptime(enrolled_in['start'], '%Y-%m-%d')
        enterprise_customer = EnterpriseCustomerFactory(name=enterprise_customer_name)
        dashboard_url = 'http://lms.example.com'
        # create a template for this customer + self enroll type (default)
        EnrollmentNotificationEmailTemplateFactory(enterprise_customer=enterprise_customer)

        if user is None:
            with raises(TypeError):
                utils.send_email_notification_message(
                    user,
                    enrolled_in,
                    dashboard_url,
                    enterprise_customer.uuid
                )
        else:
            conn = mail.get_connection()
            user_cls = user.pop('class')
            user = user_cls(**user)
            utils.send_email_notification_message(
                model_to_dict(user),
                enrolled_in,
                dashboard_url,
                enterprise_customer.uuid,
                email_connection=conn,
            )
            assert len(mail.outbox) == 1
            for field, val in expected_fields.items():
                assert getattr(
                    mail.outbox[0], field
                ) == val, f'Did not match attrib: {field} for {enterprise_customer_name}'
            assert isinstance(mail.outbox[0].connection, type(conn))

    def test_send_email_notification_with_admin_template(self):
        """
        Test that we can successfully render and send an email message using admin template
        specifically saved for a given customer.
        """
        enterprise_customer = EnterpriseCustomerFactory()
        # create a template for this customer + self enroll type (default)
        EnrollmentNotificationEmailTemplateFactory(
            enterprise_customer=enterprise_customer,
            template_type=ADMIN_ENROLL_EMAIL_TEMPLATE_TYPE,
        )

        conn = mail.get_connection()

        user = {'class': UserFactory, 'first_name': 'John', 'email': 'john@smith.com'}
        user_cls = user.pop('class')
        user = user_cls(**user)

        enrolled_in = {
            'name': 'course name',
            'url': 'lms.example.com/courses',
            'type': 'course',
            'branding': 'MicroMaster',
            'start': '2017-01-01',
        }

        dashboard_url = 'http://lms.example.com'

        utils.send_email_notification_message(
            model_to_dict(user),
            enrolled_in,
            dashboard_url,
            enterprise_customer.uuid,
            email_connection=conn,
            admin_enrollment=True,
        )
        assert len(mail.outbox) == 1
        # no need to test the entire template render content as long as it does not fail to render
        assert len(mail.outbox[0].alternatives) == 1
        assert mail.outbox[0].alternatives[0] == (
            f'<html><body>Dear John lms.example.com/courses, course name, {enterprise_customer.name}</body></html>',
            'text/html'
        )

    @ddt.data(
        (
            {'class': PendingEnterpriseCustomerUserFactory, 'user_email': 'john@smith.com'},
            {
                'name': 'Demo Course',
                'url': 'http://lms.example.com/courses',
                'type': 'course',
                'start': '2017-01-01'
            },
            'EdX',
            {
                'subject': 'New course! Demo Course!',
                'from_email': 'course_staff@example.com',
                'to': [
                    'john@smith.com'
                ],
                'body': (
                    'plaintext_value'
                ),
                'alternatives': [
                    (
                        (
                            '<b>HTML value</b>'
                        ),
                        'text/html'
                    )
                ],
                'attachments': [],
            }
        ),
        (
            {'class': UserFactory, 'first_name': 'John', 'username': '', 'email': 'john@smith.com'},
            {
                'name': 'Enterprise Learning',
                'url': 'http://lms.example.com/courses',
                'type': 'course',
                'start': '2017-06-23'
            },
            'Widgets, Inc',
            {
                'subject': 'New course! Enterprise Learning!',
                'from_email': 'course_staff@example.com',
                'to': [
                    'john@smith.com'
                ],
                'body': (
                    'plaintext_value'
                ),
                'alternatives': [
                    (
                        (
                            '<b>HTML value</b>'
                        ),
                        'text/html'
                    )
                ],
                'attachments': [],
            }
        ),
        (
            {'class': UserFactory, 'username': 'johnny_boy', 'email': 'john@smith.com', 'first_name': ''},
            {
                "name": "Master of Awesomeness",
                "url": "http://lms.example.com/courses",
                "type": "program",
                "branding": "MicroMaster",
                "start": "2017-04-15",
            },
            'MIT',
            {
                'subject': 'New course! Master of Awesomeness!',
                'from_email': 'course_staff@example.com',
                'to': [
                    'john@smith.com'
                ],
                'body': (
                    'plaintext_value'
                ),
                'alternatives': [
                    (
                        (
                            '<b>HTML value</b>'
                        ),
                        'text/html'
                    )
                ],
                'attachments': [],
            }
        ),
        (
            None,
            {
                'name': 'coursename',
                'url': 'lms.example.com/courses',
                'type': 'program',
                'branding': 'MicroMaster',
                'start': '2017-01-01',
            },
            'EdX',
            {}
        ),
    )
    @ddt.unpack
    def test_send_email_notification_message_with_site_defined_values(
            self,
            user,
            enrolled_in,
            enterprise_customer_name,
            expected_fields,
    ):
        """
        Test that we can successfully render and send an email message.
        """
        enrolled_in['start'] = datetime.datetime.strptime(enrolled_in['start'], '%Y-%m-%d')
        dashboard_url = 'http://lms.example.com'
        enterprise_customer = EnterpriseCustomerFactory(name=enterprise_customer_name)
        # since migration already sets up fallback templates, use per customer template
        EnrollmentNotificationEmailTemplateFactory(
            enterprise_customer=enterprise_customer,
            plaintext_template='plaintext_value',
            html_template='<b>HTML value</b>',
            subject_line="New course! {course_name}!",
        )
        if user is None:
            with raises(TypeError):
                utils.send_email_notification_message(
                    user,
                    enrolled_in,
                    dashboard_url,
                    enterprise_customer.uuid,
                )
        else:
            conn = mail.get_connection()
            user_cls = user.pop('class')
            user = user_cls(**user)
            utils.send_email_notification_message(
                model_to_dict(user),
                enrolled_in,
                dashboard_url,
                enterprise_customer.uuid,
                email_connection=conn,
            )
            assert len(mail.outbox) == 1
            for field, val in expected_fields.items():
                assert getattr(mail.outbox[0], field) == val
            assert isinstance(mail.outbox[0].connection, type(conn))

    @ddt.data(
        (
            'override@example.com',
            'override@example.com'
        ),
        (
            None,
            'course_staff@example.com'
        )
    )
    @ddt.unpack
    @mock.patch('enterprise.utils.get_enterprise_customer')
    def test_send_email_notification_message_with_site_from_email_override(
            self,
            site_config_from_email_address,
            expected_from_email_address,
            mock_get_ent_customer,
    ):
        """
        Test that we can successfully override a from email address per site.
        """
        user = UserFactory(username='sal', email='sal@smith.com', first_name='sal')
        enrolled_in = {
            'name': 'Demo Course',
            'url': 'http://lms.example.com/courses',
            'type': 'course',
            'start': '2017-01-01'
        }
        enrolled_in['start'] = datetime.datetime.strptime(enrolled_in['start'], '%Y-%m-%d')
        dashboard_url = 'http://lms.example.com'

        site = SiteFactory()
        if site_config_from_email_address:
            site.configuration = mock.MagicMock(
                get_value=mock.MagicMock(
                    return_value=site_config_from_email_address
                )
            )

        enterprise_customer = EnterpriseCustomerFactory(site=site)

        mock_get_ent_customer.return_value = enterprise_customer

        conn = mail.get_connection()
        utils.send_email_notification_message(
            model_to_dict(user),
            enrolled_in,
            dashboard_url,
            enterprise_customer.uuid,
            email_connection=conn,
        )

        assert len(mail.outbox) == 1
        assert mail.outbox[0].from_email == expected_from_email_address
        assert isinstance(mail.outbox[0].connection, type(conn))

    def test_get_enterprise_customer_for_user(self):
        """
        Test `get_enterprise_customer_for_user` helper method.
        """
        faker = FakerFactory.create()
        provider_id = faker.slug()  # pylint: disable=no-member

        user = UserFactory(id=1)
        ecu = EnterpriseCustomerUserFactory(
            user_id=user.id,
        )
        EnterpriseCustomerIdentityProviderFactory(
            enterprise_customer=ecu.enterprise_customer,
            provider_id=provider_id,
        )

        # Assert that correct enterprise customer is returned
        self.assertEqual(
            utils.get_enterprise_customer_for_user(auth_user=user),
            ecu.enterprise_customer,
        )

        # Assert that None is returned if user is not associated with any enterprise customer
        self.assertEqual(
            utils.get_enterprise_customer_for_user(auth_user=UserFactory(id=2)),
            None,
        )

    def test_get_enterprise_customer_uuid_for_user_and_course(self):
        """
        Test `get_enterprise_customer_for_user_and_course` helper method.
        """
        user = UserFactory(id=1)
        ece1 = EnterpriseCourseEnrollmentFactory(enterprise_customer_user__user_id=user.id)
        course_run_id = ece1.course_id
        expected_uuid1 = str(ece1.enterprise_customer_user.enterprise_customer.uuid)

        # Assert that 1 enrollments gives you correct list.
        self.assertEqual(
            utils.get_enterprise_uuids_for_user_and_course(user, course_run_id),
            [expected_uuid1],
        )

        ece2 = EnterpriseCourseEnrollmentFactory(
            course_id=course_run_id,
            enterprise_customer_user__user_id=user.id)
        expected_uuid2 = str(ece2.enterprise_customer_user.enterprise_customer.uuid)
        # Assert that 2 enrollments gives you 2 uuids
        uuid_list = utils.get_enterprise_uuids_for_user_and_course(user, course_run_id)
        self.assertIn(expected_uuid1, uuid_list)
        self.assertIn(expected_uuid2, uuid_list)
        self.assertEqual(2, len(uuid_list))

        # Assert that a customer from an unrelated enrollment is not returned:
        ece3 = EnterpriseCourseEnrollmentFactory(enterprise_customer_user__user_id=user.id)
        expected_uuid3 = str(ece3.enterprise_customer_user.enterprise_customer.uuid)
        self.assertNotIn(expected_uuid3, utils.get_enterprise_uuids_for_user_and_course(user, course_run_id))

        # Test that active filter query param works:
        ece4 = EnterpriseCourseEnrollmentFactory(
            enterprise_customer_user__user_id=user.id,
            course_id=course_run_id,
            enterprise_customer_user__enterprise_customer__active=False
        )
        expected_uuid4 = str(ece4.enterprise_customer_user.enterprise_customer.uuid)
        self.assertNotIn(expected_uuid4, utils.get_enterprise_uuids_for_user_and_course(
            user,
            course_run_id,
            is_customer_active=True))
        self.assertIn(expected_uuid4, utils.get_enterprise_uuids_for_user_and_course(
            user,
            course_run_id))
        self.assertIn(expected_uuid4, utils.get_enterprise_uuids_for_user_and_course(
            user,
            course_run_id,
            is_customer_active=False))

    @ddt.data(
        (
            'lms.example.com/courses/course-v1:edx+test-course+T22017/',
            {},
            ['lms.example.com/courses/course-v1:edx+test-course+T22017/'],
        ),
        (
            'http://localhost/courses/course-v1:edx+test-course+T22017/',
            {},
            ['http://localhost/courses/course-v1:edx+test-course+T22017/'],
        ),
        (
            'http://open.edx/courses/course-v1:edx+test-course+T22017/?course=test-course',
            {},
            [
                'http://open.edx/courses/course-v1:edx+test-course+T22017/?',
                'course=test-course',
            ],
        ),
        (
            'http://open.edx/courses/course-v1:edx+test-course+T22017/',
            {
                'tpa_hint': 'test-shib',
                'referrer': 'edX',
            },
            [
                'http://open.edx/courses/course-v1:edx+test-course+T22017/?',
                'tpa_hint=test-shib',
                'referrer=edX',
            ],
        ),
        (
            'http://open.edx/courses/course-v1:edx+test-course+T22017/?course=test-course',
            {
                'tpa_hint': 'test-shib',
                'referrer': 'edX',
            },
            [
                'http://open.edx/courses/course-v1:edx+test-course+T22017/?',
                'tpa_hint=test-shib',
                'referrer=edX',
                'course=test-course',
            ],
        ),
    )
    @ddt.unpack
    def test_update_query_parameters(self, url, query_parameters, expected_url_parts):
        """
        Test `update_query_parameters` helper method.
        """
        url = utils.update_query_parameters(url, query_parameters)

        # Make sure all query parameters are present We can not assert whole urls because we can
        # not determine the position of these query parameters.
        # e.g. `http://<base>?course=test-course&tpa_hint=test-shib` and
        # `http://<base>?tpa_hint=test-shib&course=test-course` are both same urls but different strings.
        for url_part in expected_url_parts:
            assert url_part in url

    @ddt.data(
        (
            {
                'key': 'course-v1:edx+test-course+T22017',
                'uuid': '57432370-0a6e-4d95-90fe-77b4fe64de2b',
                'title': 'A self-paced audit course',
            },
            {},
            ['http://testserver/course_modes/choose/course-v1:edx+test-course+T22017/'],
        ),
        (
            {
                'key': 'course-v1:edx+test-course+T22017',
                'uuid': '57432370-0a6e-4d95-90fe-77b4fe64de2b',
                'title': 'A self-paced audit course',
            },
            {
                'tpa_hint': 'test-shib',
                'referrer': 'edX',
            },
            [
                'http://testserver/course_modes/choose/course-v1:edx+test-course+T22017/?',
                'tpa_hint=test-shib',
                'referrer=edX',
            ],
        ),
    )
    @ddt.unpack
    @override_settings(LMS_ROOT_URL='http://testserver/')
    def test_get_course_track_selection_url(self, course_run, query_parameters, expected_url_parts):
        """
        Test `get_course_track_selection_url` helper method.
        """
        course_root = "course_modes/choose/{course_id}/".format(course_id=course_run.get('key', ''))

        with mock.patch('enterprise.utils.reverse', return_value=course_root):
            url = utils.get_course_track_selection_url(course_run, query_parameters)

            # Make sure course run url returned by get_course_track_selection_url
            # contains all the expected url parts.
            # We can not assert whole urls because we can not determine the position of these query parameters.
            # e.g. `http://<base>?course=test-course&tpa_hint=test-shib` and
            # `http://<base>?tpa_hint=test-shib&course=test-course` are both same urls but different strings.
            for url_part in expected_url_parts:
                assert url_part in url

    def test_get_course_track_selection_url_raises_exception(self):
        """
        Test `get_course_track_selection_url` raises exception for missing `key` in course run.
        """
        with raises(KeyError):
            utils.get_course_track_selection_url({}, {})

    @ddt.data(
        ('{} hour', '{} hours', '{}-{} hours', 2, 4, '2-4 hours'),
        ('{} hour', '{} hours', '{}-{} hours', 2, 2, '2 hours'),
        ('{} hour', '{} hours', '{}-{} hours', 1, 1, '1 hour'),
        ('{} hour', '{} hours', '{}-{} hours', None, 1, '1 hour'),
        ('{} hour', '{} hours', '{}-{} hours', 1, None, '1 hour'),
        ('{} hour', '{} hours', '{}-{} hours', None, 3, '3 hours'),
        ('{} hour', '{} hours', '{}-{} hours', 3, None, '3 hours'),
        ('{} hour', '{} hours', '{}-{} hours', None, None, None),
    )
    @ddt.unpack
    def test_ungettext_min_max(self, singular, plural, range_text, min_val, max_val, expected_output):
        """
        ``ungettext_min_max`` returns the appropriate strings depending on a certain min & max.
        """
        assert utils.ungettext_min_max(singular, plural, range_text, min_val, max_val) == expected_output

    @mock.patch('enterprise.utils.configuration_helpers')
    def test_get_configuration_value_with_openedx(self, config_mock):
        """
        ``get_configuration_value`` returns the appropriate non-default value when connected to Open edX.
        """
        config_mock.get_value.return_value = 'value'
        assert utils.get_configuration_value('value', default='default') == 'value'

    @mock.patch('enterprise.utils.get_url')
    def test_get_configuration_value_url_type(self, get_url_mock):
        """
        ``get_configuration_value`` returns the appropriate non-default value for URL types when in Open edX.
        """
        get_url_mock.return_value = 'value'
        assert utils.get_configuration_value('value', default='default', type='url') == 'value'

    def test_get_configuration_value_without_openedx(self):
        """
        ``get_configuration_value`` returns a default value of 'default' when not connected to Open edX.
        """
        assert utils.get_configuration_value('value', default='default') == 'default'

    def test_get_configuration_value_for_site_with_configuration(self):
        """
        ``get_configuration_value_for_site`` returns the key's value or the default in the site configuration.

        We do not test whether we get back the key's value or the default in particular, but just that
        the function returns a value through the configuration, rather than the default.
        """
        site = SiteFactory()
        site.configuration = mock.MagicMock(get_value=mock.MagicMock(return_value='value'))
        assert utils.get_configuration_value_for_site(site, 'key', 'default') == 'value'

    def test_get_configuration_value_for_site_without_configuration(self):
        """
        ``get_configuration_value_for_site`` returns the default because of no site configuration.
        """
        assert utils.get_configuration_value_for_site(SiteFactory(), 'key', 'default') == 'default'

    @ddt.data('GET', 'DELETE')
    def test_get_request_value_query_params(self, method):
        """
        Request value is retrieved from the query parameters and not posted data for certain methods.
        """
        request = mock.MagicMock(method=method, query_params={'key': 'query_params'}, data={'key': 'data'})
        assert utils.get_request_value(request, 'key') == 'query_params'

    def test_get_request_value_data(self):
        """
        Request value is retrieved from the posted data and not from the query parameters for certain methods.
        """
        request = mock.MagicMock(method='POST', query_params={'key': 'query_params'}, data={'key': 'data'})
        assert utils.get_request_value(request, 'key') == 'data'

    @ddt.data(
        (
            'MicroMasters Certificate',
            'A series of Master’s-level courses to advance your career, '
            'created by top universities and recognized by companies. '
            'MicroMasters Programs are credit-eligible, provide in-demand '
            'knowledge and may be applied to accelerate a Master’s Degree.',
        ),
        (
            'Professional Certificate',
            'Designed by industry leaders and top universities to enhance '
            'professional skills, Professional Certificates develop the '
            'proficiency and expertise that employers are looking for with '
            'specialized training and professional education.',
        ),
        (
            'XSeries Certificate',
            'Created by world-renowned experts and top universities, XSeries '
            'are designed to provide a deep understanding of key subjects '
            'through a series of courses. Complete the series to earn a valuable '
            'XSeries Certificate that illustrates your achievement.',
        ),
        (
            'Random Certificate',
            '',
        )
    )
    @ddt.unpack
    def test_get_program_type_description(self, program_type, expected_description):
        """
        ``get_program_type_description`` should return the appropriate description for any program type.
        """
        assert utils.get_program_type_description(program_type) == expected_description

    @mock.patch('enterprise.utils.segment')
    def test_track_event(self, analytics_mock):
        """
        ``track_event`` fires a track event to segment.
        """
        utils.track_event('user_id', 'event_name', 'properties')
        analytics_mock.track.assert_called_once()

    @mock.patch('enterprise.utils.track_event')
    def test_track_enrollment(self, track_event_mock):
        """
        ``track_enrollment`` invokes ``track_event`` with custom properties.
        """
        pathway = 'some-pathway'
        user_id = 123123
        course_run_id = 'course-v1:Some+edX+Course'
        url_path = '/some/url/path'
        utils.track_enrollment(pathway, user_id, course_run_id, url_path)
        track_event_mock.assert_called_once_with(user_id, 'edx.bi.user.enterprise.onboarding', {
            'pathway': pathway,
            'url_path': url_path,
            'course_run_id': course_run_id,
        })

    @ddt.data(
        ("2014-10-13T13:11:03Z", utils.parse_datetime("2014-10-13T13:11:03Z")),
        (datetime.datetime(2020, 5, 17), datetime.datetime(2020, 5, 17).replace(tzinfo=pytz.UTC)),
        (None, None)
    )
    @ddt.unpack
    def test_parse_datetime_handle_invalid(self, datetime_value, expected_parsed_datetime):
        """
        ``parse_datetime_handle_invalid`` wraps ``parse_datetime`` such that it returns ``None`` for any bad types.
        """
        assert utils.parse_datetime_handle_invalid(datetime_value) == expected_parsed_datetime

    @ddt.data(
        (
            # advertised course run is set and included, return that one
            {
                "advertised_course_run_uuid": "dd7bb3e4-56e9-4639-9296-ea9c2fb99c7f",
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(uuid='dd7bb3e4-56e9-4639-9296-ea9c2fb99c7f'),
                    fake_catalog_api.create_course_run_dict(uuid='28e2d4c2-a020-4959-b461-6f879bbe1391')
                ]
            },
            fake_catalog_api.create_course_run_dict(uuid='dd7bb3e4-56e9-4639-9296-ea9c2fb99c7f')
        ),
        (
            # advertised course run is set but not included, return None
            {
                "advertised_course_run_uuid": "dd7bb3e4-56e9-4639-9296-ea9c2fb99c7f",
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(uuid='28e2d4c2-a020-4959-b461-6f879bbe1391')
                ]
            },
            None
        ),
        (
            # advertised course run is not set, return None
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(uuid='28e2d4c2-a020-4959-b461-6f879bbe1391')
                ]
            },
            None
        ),
    )
    @ddt.unpack
    def test_get_advertised_course_run(self, course, expected_course_run):
        assert utils.get_advertised_course_run(course) == expected_course_run

    @ddt.data(
        (
            # everything true, course is active
            fake_catalog_api.create_course_run_dict(status='published', is_enrollable=True, is_marketable=True),
            True
        ),
        (
            # not published
            fake_catalog_api.create_course_run_dict(status='unpublished', is_enrollable=True, is_marketable=True),
            False
        ),
        (
            # not enrollable
            fake_catalog_api.create_course_run_dict(status='published', is_enrollable=False, is_marketable=True),
            False
        ),
        (
            # not marketable
            fake_catalog_api.create_course_run_dict(status='published', is_enrollable=True, is_marketable=False),
            False
        )
    )
    @ddt.unpack
    def test_is_course_run_active(self, course_run, expected_result):
        assert utils.is_course_run_active(course_run) == expected_result

    @ddt.data(
        (
            {
                "min_effort": 2,
                "max_effort": 3,
                "weeks_to_complete": 3,
            },
            '2-3 hours a week for 3 weeks. ',
        ),
        (
            {
                "min_effort": 0,
                "max_effort": 3,
                "weeks_to_complete": 3,
            },
            '',
        ),
        (
            {
                "min_effort": 2,
                "max_effort": '',
                "weeks_to_complete": 3,
            },
            '',
        ),
        (
            {
                "min_effort": 2,
                "max_effort": 2,
                "weeks_to_complete": None,
            },
            '',
        ),
        (
            {
                "min_effort": '',
                "max_effort": 0,
                "weeks_to_complete": None,
            },
            '',
        ),

    )
    @ddt.unpack
    def test_get_course_run_duration_info(self, course_run, expected_duration_info):
        """
        ``get_course_run_duration_info`` returns the course run duration info
        """
        assert utils.get_course_run_duration_info(course_run) == expected_duration_info

    @ddt.data(
        (
            {
                "status": "published",
                "end": "3000-10-13T13:11:01Z",
                "enrollment_start": "2014-10-13T13:11:03Z",
                "enrollment_end": "2999-10-13T13:11:04Z",
            },
            True,
        ),
        (
            {
                "status": "published",
                "end": None,
                "enrollment_start": "2014-10-13T13:11:03Z",
                "enrollment_end": "2999-10-13T13:11:04Z"
            },
            True,
        ),
        (
            {
                "status": "published",
                "end": "3000-10-13T13:11:01Z",
                "enrollment_start": None,
                "enrollment_end": "2999-10-13T13:11:04Z"
            },
            True,
        ),
        (
            {
                "status": "published",
                "end": "3000-10-13T13:11:01Z",
                "enrollment_start": "2014-10-13T13:11:03Z",
                "enrollment_end": None
            },
            True,
        ),
        (
            {
                "status": "published",
                "end": "2014-10-13T13:11:01Z",  # end < now
                "enrollment_start": "2014-10-13T13:11:03Z",
                "enrollment_end": "2999-10-13T13:11:04Z",
            },
            False,
        ),
        (
            {
                "status": "published",
                "end": "3000-10-13T13:11:01Z",
                "enrollment_start": "2999-10-13T13:11:03Z",  # enrollment_start > now
                "enrollment_end": "3000-10-13T13:11:04Z",
            },
            False,
        ),
        (
            {
                "status": "published",
                "end": "3000-10-13T13:11:01Z",
                "enrollment_start": "2014-10-13T13:11:03Z",
                "enrollment_end": "2014-10-13T13:11:04Z",  # enrollment_end < now
            },
            False,
        ),
        (
            {
                "status": "unpublished",
                "end": "3000-10-13T13:11:01Z",
                "enrollment_start": "2014-10-13T13:11:03Z",
                "enrollment_end": "3000-10-13T13:11:04Z",  # enrollment_end < now
            },
            False,
        ),
    )
    @ddt.unpack
    def test_is_course_run_enrollable(self, course_run, expected_enrollment_eligibility):
        """
        ``is_course_run_enrollable`` returns whether the course run is enrollable.

        We check that the function returns:
        - False on end < now.
        - False on enrollment_end < now.
        - False on enrollment_start > now.
        - False on is_course_run_published = False
        - True if none of the above plus, optionally, if any of the values are NULL.
        """
        assert utils.is_course_run_enrollable(course_run) == expected_enrollment_eligibility

    @ddt.data(
        (
            [],
            False,
        ),
        (
            [
                {
                    "status": "published",
                    "availability": "Current",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2999-10-13T13:11:03Z",  # enrollment_start > now
                    "enrollment_end": "3000-10-13T13:11:04Z",
                },
                {
                    "status": "published",
                    "availability": "Current",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                },
            ],
            True,
        ),
        (
            [
                {
                    "status": "published",
                    "availability": "Archived",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                },
                {
                    "status": "published",
                    "availability": "Current",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                },
            ],
            True,
        ),
        (
            [
                {
                    "status": "published",
                    "availability": "Current",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                }
            ],
            True,
        ),
        (
            [
                {
                    "status": "published",
                    "availability": "Archived",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                }
            ],
            False,
        ),
        (
            [
                {
                    "status": "published",
                    "availability": "Current",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                }
            ],
            True,
        ),
        (
            [
                {
                    "status": "unpublished",
                    "availability": "Current",
                    "end": "3000-10-13T13:11:01Z",
                    "enrollment_start": "2014-10-13T13:11:03Z",
                    "enrollment_end": "2999-10-13T13:11:04Z",
                }
            ],
            False,
        ),
    )
    @ddt.unpack
    def test_contains_course_run_available_for_enrollment(self, course_runs, expected_enrollment_eligibility):
        """
        tests contains_course_run_available_for_enrollment functionality
        """
        assert utils.has_course_run_available_for_enrollment(course_runs) == expected_enrollment_eligibility

    @ddt.data(
        (
            [],
            None,
        ),
        (
            None,
            None,
        ),
        (
            [
                {
                    "end": "2020-09-13T13:11:01Z",
                },
                {
                    "end": "2020-10-13T04:45:00Z",
                },
                {
                    "end": "2020-10-13T13:11:15Z",
                },
                {
                    "end": "2019-07-13T01:11:01Z",
                },
            ],
            "2020-10-13T13:11:15Z",
        ),
        (
            [
                {
                    "end": "3000-05-29T12:11:01Z",
                },
                {
                    "end": "3000-10-13T13:11:01Z",
                },
            ],
            "3000-10-13T13:11:01Z",  # future dates
        ),
        (
            [
                {
                    "end": "3000-05-29T12:11:01Z",
                },
                {
                    "end": "2020-05-29T12:11:01Z",
                },
                {
                    "end": None,
                },
            ],
            "3000-05-29T12:11:01Z",  # one of the end date is None
        ),
        (
            [
                {
                    "end": None,
                },
            ],
            None,
        ),
        (
            [
                {
                    "end": None,
                },
                {
                    "end": None,
                },
                {
                    "end": None,
                },
            ],
            None,  # all of the end dates are None
        ),
        (
            [
                {
                    "end": "3000-05-29T12:11:01Z",
                },
                {
                    "end": "3000-05-29T12:11:01Z",
                },
            ],
            "3000-05-29T12:11:01Z",  # end dates fall on same date
        ),
    )
    @ddt.unpack
    def test_course_end_date(self, course_runs, expected_end_date):
        """
        Tests that course end_date is always equal to latest course run's end date.
        """
        assert utils.get_last_course_run_end_date(course_runs) == expected_end_date

    @ddt.data(
        ({"seats": [{"type": "verified", "upgrade_deadline": "3000-10-13T13:11:04Z"}]}, True),
        ({"seats": [{"type": "verified", "upgrade_deadline": None}]}, True),
        ({"seats": [{"type": "verified"}]}, True),
        ({"seats": [{"type": "verified", "upgrade_deadline": "1977-10-13T13:11:04Z"}]}, False),
        ({"seats": [{"type": "audit"}]}, False),
        ({}, False),
    )
    @ddt.unpack
    def test_is_course_run_upgradeable(self, course_run, expected_upgradeability):
        """
        ``is_course_run_upgradeable`` returns whether the course run is eligible
        for an upgrade to the verified track.
        """
        assert utils.is_course_run_upgradeable(course_run) == expected_upgradeability

    @ddt.data(
        ({"status": "published"}, True),
        ({"status": "unpublished"}, False)
    )
    @ddt.unpack
    def test_is_course_published(self, course_run, expected_response):
        """
        ``is_course_run_published`` returns whether the course run is considered "published"
        given its metadata structure and values.
        """
        assert utils.is_course_run_published(course_run) == expected_response

    @ddt.data(
        ({"start": "3000-10-13T13:11:04Z"}, utils.parse_datetime_handle_invalid("3000-10-13T13:11:04Z")),
        ({}, None),
    )
    @ddt.unpack
    def test_get_course_run_start(self, course_run, expected_start):
        """
        ``get_course_run_start`` returns the start date given the course run dict.
        """
        assert utils.get_course_run_start(course_run) == expected_start

    def test_get_course_run_start_with_default(self):
        """
        ``get_course_run_start`` returns the appropriate default start date.
        """
        now = datetime.datetime.now()
        assert utils.get_course_run_start({}) is None
        assert utils.get_course_run_start({}, now) == now

    @ddt.data(
        (
            fake_catalog_api.create_course_run_dict(
                end=DATETIME_NOW + datetime.timedelta(days=20),
                weeks_to_complete=5
            ),
            True,
        ),
        (
            fake_catalog_api.create_course_run_dict(
                end=DATETIME_NOW + datetime.timedelta(days=20),
                weeks_to_complete=2
            ),
            False,
        ),
        (
            fake_catalog_api.create_course_run_dict(
                end=DATETIME_NOW + datetime.timedelta(days=20),
                weeks_to_complete=None
            ),
            False,
        ),
    )
    @ddt.unpack
    def test_is_course_run_about_to_end(self, course_run, expected_boolean):
        """
        ``is_course_run_about_to_end`` returns the boolean is course_run about to end.
        """
        assert utils.is_course_run_about_to_end(course_run) == expected_boolean

    @ddt.data(
        (
            # Test with two Current runs, but one not published. Should choose published one.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(
                        availability="Current",
                        start="2015-10-15T13:11:03Z",
                        status="unpublished",
                        end="2014-10-15T13:11:03Z",
                        enrollment_start=None,
                        enrollment_end=None,
                    ),
                    fake_catalog_api.create_course_run_dict(
                        availability="Current",
                        start="2015-10-15T13:11:03Z",
                        end="2015-10-15T13:11:03Z",
                        enrollment_start=None,
                        enrollment_end=None,
                    ),
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(
                availability="Current",
                start="2015-10-15T13:11:03Z",
                end="2015-10-15T13:11:03Z",
                enrollment_start=None,
                enrollment_end=None,
            ),
        ),
        (
            # Test with two enrollable/upgradeable course runs.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(),
                    fake_catalog_api.create_course_run_dict(start="2014-10-15T13:11:03Z"),
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(start="2014-10-15T13:11:03Z"),
        ),
        (
            # Test with one enrollable course run.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(),
                    fake_catalog_api.create_course_run_dict(
                        enrollment_end="2014-10-14T13:11:03Z",
                    ),
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(),
        ),
        (
            # Test with no enrollable course runs.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(enrollment_end="2014-10-14T13:11:03Z"),
                    fake_catalog_api.create_course_run_dict(
                        start="2014-10-15T13:11:03Z",
                        enrollment_end="2014-10-14T13:11:03Z",
                    ),
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(
                start="2014-10-15T13:11:03Z",
                enrollment_end="2014-10-14T13:11:03Z",
            ),
        ),
        (
            # Test with one upgradeable course run.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(),
                    fake_catalog_api.create_course_run_dict(
                        upgrade_deadline="2014-10-14T13:11:03Z",
                    ),
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(),
        ),
        (
            # Test with no upgradeable course runs.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(upgrade_deadline="2014-10-14T13:11:03Z"),
                    fake_catalog_api.create_course_run_dict(
                        start="2014-10-15T13:11:03Z",
                        upgrade_deadline="2014-10-14T13:11:03Z",
                    ),
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(
                start="2014-10-15T13:11:03Z",
                upgrade_deadline="2014-10-14T13:11:03Z",
            ),
        ),
        (
            # Test with current availability.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(
                        end=DATETIME_NOW + datetime.timedelta(days=20),
                        availability='Current'
                    ),
                    fake_catalog_api.create_course_run_dict(end="2099-01-14T13:11:03Z")
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(
                end=DATETIME_NOW + datetime.timedelta(days=20),
                availability='Current'
            ),
        ),
        (
            # Test with current availability.
            {
                "course_runs": [
                    fake_catalog_api.create_course_run_dict(
                        end=DATETIME_NOW + datetime.timedelta(days=20),
                        weeks_to_complete=4,
                        availability='Current',
                    ),
                    fake_catalog_api.create_course_run_dict()
                ],
            },
            [],
            fake_catalog_api.create_course_run_dict(),
        ),
        (   # It will return the active run
            {
                "course_runs": [],
            },
            [fake_catalog_api.create_course_run_dict(start="2014-10-15T13:11:03Z")],
            fake_catalog_api.create_course_run_dict(start="2014-10-15T13:11:03Z"),
        ),
        (   # it will return the closest active run
            {
                "course_runs": [],
            },
            [
                fake_catalog_api.create_course_run_dict(start="2014-10-15T13:11:03Z"),
                fake_catalog_api.create_course_run_dict(start="2019-10-15T13:11:03Z"),
            ],
            fake_catalog_api.create_course_run_dict(start="2019-10-15T13:11:03Z"),
        ),
        (
            # Test with no course runs.
            {
                "course_runs": [],
            },
            [],
            None
        ),
    )
    @ddt.unpack
    def test_get_current_course_run(self, course, users_active_course_runs, expected_course_run):
        """
        ``get_current_course_run`` returns the current course run for the given course dictionary.
        """
        assert utils.get_current_course_run(course, users_active_course_runs) == expected_course_run

    @ddt.data(
        (
            [
                fake_catalog_api.create_course_run_dict(start="2019-10-15T13:11:03Z"),
                fake_catalog_api.create_course_run_dict(start="2014-10-15T13:11:03Z"),
            ],
            fake_catalog_api.create_course_run_dict(start="2019-10-15T13:11:03Z"),
        ),
        (
            [
                fake_catalog_api.create_course_run_dict(start="2019-10-15T13:11:03Z"),
            ],
            fake_catalog_api.create_course_run_dict(start="2019-10-15T13:11:03Z"),
        ),

    )
    @ddt.unpack
    def test_get_closest_course_run(self, course_runs, expected_course_run):
        """
        ``get_closest_course_run`` returns the closest course run.
        """
        assert utils.get_closest_course_run(course_runs) == expected_course_run

    @ddt.data(
        (
            {
                "course_runs":
                    [
                        {'key': 'fake-key1'},
                        {'key': 'fake-key2'}
                    ],
            },
            [
                {
                    'is_active': True,
                    'course_details': {'course_id': 'fake-key1'}
                },
                {
                    'is_active': False,
                    'course_details': {'course_id': 'fake-key2'}
                }
            ],
            [
                {'key': 'fake-key1'},
            ],
        ),
        (
            {
                "course_runs":
                    [
                        {'key': 'fake-key1'},
                        {'key': 'fake-key2'}
                    ],
            },
            [
                {
                    'is_active': True,
                    'course_details': {'course_id': 'fake-key1'}
                },
                {
                    'is_active': True,
                    'course_details': {'course_id': 'fake-key2'}
                }
            ],
            [
                {'key': 'fake-key1'},
                {'key': 'fake-key2'}
            ],
        ),
        (
            {
                "course_runs":
                    [
                        {'key': 'fake-key1'},
                        {'key': 'fake-key2'}
                    ],
            },
            [
                {
                    'is_active': False,
                    'course_details': {'course_id': 'fake-key1'}
                },
                {
                    'is_active': False,
                    'course_details': {'course_id': 'fake-key2'}
                }
            ],
            [],
        ),
        (
            {
                "course_runs": [],
            },
            [
                {
                    'is_active': True,
                    'course_details': {'course_id': 'fake-key1'}
                }
            ],
            [],
        ),

    )
    @ddt.unpack
    def test_get_active_course_runs(self, course, users_all_enrolled_courses, expected_course_run):
        """
        ``get_active_course_runs`` returns active course runs of given course's course runs.
        """
        assert utils.get_active_course_runs(course, users_all_enrolled_courses) == expected_course_run

    @mock.patch('enterprise.utils.Registry')
    def test_get_idp_choices(self, mock_registry):
        Provider = namedtuple('SAMLProvider', 'provider_id, name, disable_for_enterprise_sso')
        enabled_provider = Provider('12345', 'google', False)
        disabled_provider = Provider('333333', 'facebook', True)

        mock_registry.enabled.return_value = [enabled_provider, disabled_provider]
        assert utils.get_idp_choices() == [("", "-" * 7)] + [(enabled_provider.provider_id, enabled_provider.name)]


@mark.django_db
@ddt.ddt
class TestSAPSuccessFactorsUtils(unittest.TestCase):
    """
    Tests for sap success factors utility functions.
    """

    def setUp(self):
        """
        Set up test environment.
        """
        super().setUp()
        faker = FakerFactory.create()
        self.user = UserFactory()
        self.uuid = faker.uuid4()  # pylint: disable=no-member
        self.customer = EnterpriseCustomerFactory(uuid=self.uuid)
        self.enterprise_configuration = SAPSuccessFactorsEnterpriseCustomerConfiguration(
            enterprise_customer=self.customer,
            sapsf_base_url='enterprise.successfactors.com',
            decrypted_key='key',
            decrypted_secret='secret',
        )

    @ddt.data(
        (
            {'mode': 'enroll', 'number': 0},
            {'mode': 'audit', 'number': 1},
            {'mode': 'another_audit', 'number': 2},
            {'mode': 'enroll', 'number': 3},
            {'mode': 'audit', 'number': 4},
        )
    )
    @override_settings(ENTERPRISE_COURSE_ENROLLMENT_AUDIT_MODES=['audit', 'another_audit'])
    def test_filter_audit_course_modes(self, course_modes):
        # when the audit enrollment flag is disabled
        self.customer.enable_audit_enrollment = False
        # course modes are filtered out if their mode is in the ENTERPRISE_COURSE_ENROLLMENT_AUDIT_MODES setting
        filtered_course_modes = utils.filter_audit_course_modes(self.customer, course_modes)
        assert len(filtered_course_modes) == 2
        result = [course_mode['number'] for course_mode in filtered_course_modes]
        expected = [0, 3]
        self.assertListEqual(result, expected)

        # when the audit enrollment flag is enabled
        self.customer.enable_audit_enrollment = True
        # audit course modes are not filtered out
        filtered_course_modes = utils.filter_audit_course_modes(self.customer, course_modes)
        assert len(filtered_course_modes) == 5


@mark.django_db
@ddt.ddt
class TestValidateEmailToLink(unittest.TestCase):
    """
    Tests for :method:`validate_email_to_link`.
    """

    def setUp(self):
        """
        Set up test environment.
        """
        super().setUp()
        self.enterprise_customer = EnterpriseCustomerFactory()

    def test_validate_email_to_link_normal(self):
        email = FAKER.email()  # pylint: disable=no-member
        assert PendingEnterpriseCustomerUser.objects.filter(user_email=email).count() == 0, \
            "Precondition check - should not have PendingEnterpriseCustomerUser"
        assert EnterpriseCustomerUser.objects.count() == 0, \
            "Precondition check - should not have EnterpriseCustomerUser"

        exists = utils.validate_email_to_link(email, self.enterprise_customer)  # should not raise any Exceptions
        assert exists is False

    @ddt.unpack
    @ddt.data(
        ("something", "something", utils.ValidationMessages.INVALID_EMAIL),
        ("something_again@", "something_else", utils.ValidationMessages.INVALID_EMAIL)
    )
    def test_validate_email_to_link_invalid_email(self, email, raw_email, msg_template):
        assert EnterpriseCustomerUser.objects.get_link_by_email(email, self.enterprise_customer) is None, \
            "Precondition check - should not have EnterpriseCustomerUser or PendingEnterpriseCustomerUser"

        expected_message = msg_template.format(argument=raw_email)

        with raises(ValidationError, match=expected_message):
            utils.validate_email_to_link(email, self.enterprise_customer, raw_email)

    def test_validate_email_to_link_existing_user_record_different_enterprise(self):
        user = UserFactory()
        email = user.email
        different_enterprise = EnterpriseCustomerFactory()
        existing_record = EnterpriseCustomerUserFactory(user_id=user.id, enterprise_customer=different_enterprise)
        assert not PendingEnterpriseCustomerUser.objects.exists(), \
            "Precondition check - should not have PendingEnterpriseCustomerUser"
        assert EnterpriseCustomerUser.objects.get(user_id=user.id) == existing_record, \
            "Precondition check - should have EnterpriseCustomerUser"

        exists = utils.validate_email_to_link(email, self.enterprise_customer)
        assert not exists

    @ddt.data(True, False)
    def test_validate_email_to_link_existing_user_record_same_enterprise(self, raise_exception):
        user = UserFactory()
        email = user.email
        existing_record = EnterpriseCustomerUserFactory(user_id=user.id, enterprise_customer=self.enterprise_customer)
        assert not PendingEnterpriseCustomerUser.objects.exists(), \
            "Precondition check - should not have PendingEnterpriseCustomerUser"
        assert EnterpriseCustomerUser.objects.get(user_id=user.id) == existing_record, \
            "Precondition check - should have EnterpriseCustomerUser"

        if raise_exception:
            expected_message = utils.ValidationMessages.USER_ALREADY_REGISTERED.format(
                email=email,
                ec_name=existing_record.enterprise_customer.name,
            )

            with raises(ValidationError, match=expected_message):
                utils.validate_email_to_link(email, self.enterprise_customer)
        else:
            exists = utils.validate_email_to_link(email, self.enterprise_customer, raise_exception=False)
            assert exists

    def test_validate_email_to_link_existing_pending_record_different_enterprise(self):
        email = FAKER.email()  # pylint: disable=no-member
        different_enterprise = EnterpriseCustomerFactory()
        existing_record = PendingEnterpriseCustomerUserFactory(
            user_email=email,
            enterprise_customer=different_enterprise,
        )
        assert PendingEnterpriseCustomerUser.objects.get(
            user_email=email, enterprise_customer=existing_record.enterprise_customer,
        ) == existing_record, "Precondition check - should have PendingEnterpriseCustomerUser"
        assert not EnterpriseCustomerUser.objects.exists(), \
            "Precondition check - should not have EnterpriseCustomerUser"

        exists = utils.validate_email_to_link(email, self.enterprise_customer)
        assert not exists

    @ddt.data(True, False)
    def test_validate_email_to_link_existing_pending_record_same_enterprise(self, raise_exception):
        email = FAKER.email()  # pylint: disable=no-member
        existing_record = PendingEnterpriseCustomerUserFactory(
            user_email=email,
            enterprise_customer=self.enterprise_customer,
        )
        assert PendingEnterpriseCustomerUser.objects.get(
            user_email=email, enterprise_customer=existing_record.enterprise_customer,
        ) == existing_record, "Precondition check - should have PendingEnterpriseCustomerUser"
        assert not EnterpriseCustomerUser.objects.exists(), \
            "Precondition check - should not have EnterpriseCustomerUser"

        if raise_exception:
            expected_message = utils.ValidationMessages.USER_ALREADY_REGISTERED.format(
                email=email,
                ec_name=existing_record.enterprise_customer.name,
            )

            with raises(ValidationError, match=expected_message):
                utils.validate_email_to_link(email, self.enterprise_customer)
        else:
            exists = utils.validate_email_to_link(email, self.enterprise_customer, raise_exception=False)
            assert exists

    @ddt.data(
        (True, 'Invalid PGP Key'),
        (True, None),
        (False, TEST_PGP_KEY),
    )
    @ddt.unpack
    def test_validate_pgp_key(self, raise_exception, key):
        """
        Test the behavior of validate_pgp_key.
        """
        if raise_exception:
            with raises(ValidationError, match='Invalid PGP Key provided.'):
                validators.validate_pgp_key(key)
        else:
            validators.validate_pgp_key(key)



# Generated by Django 4.2 on 2023-12-29 17:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0197_auto_20231130_2239'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='enterprisecourseenrollment',
            options={'ordering': ['id']},
        ),
    ]



"""
Views for the ``enterprise-user`` API endpoint.
"""

from collections import OrderedDict

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, permissions, response, status
from rest_framework.pagination import PageNumberPagination

from django.contrib import auth
from django.core.exceptions import ValidationError
from django.db.models import Case, CharField, F, Q, When

from enterprise import models
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadOnlyModelViewSet
from enterprise.logging import getEnterpriseLogger

User = auth.get_user_model()

LOGGER = getEnterpriseLogger(__name__)


class EnterpriseCustomerSupportPaginator(PageNumberPagination):
    """Custom paginator for the enterprise customer support."""

    page_size = 8

    def get_paginated_response(self, data):
        """Return a paginated style `Response` object for the given output data."""
        return response.Response(
            OrderedDict(
                [
                    ("count", self.page.paginator.count),
                    ("num_pages", self.page.paginator.num_pages),
                    ("next", self.get_next_link()),
                    ("previous", self.get_previous_link()),
                    ("results", data),
                ]
            )
        )

    def paginate_queryset(self, queryset, request, view=None):
        """
        Paginate a queryset if required, either returning a page object,
        or `None` if pagination is not configured for this view.

        """
        if isinstance(queryset, filter):
            queryset = list(queryset)

        return super().paginate_queryset(queryset, request, view)


class EnterpriseCustomerSupportViewSet(EnterpriseReadOnlyModelViewSet):
    """
    API views for the ``enterprise-customer-support`` API endpoint.
    """

    queryset = models.PendingEnterpriseCustomerUser.objects.all()
    filter_backends = (DjangoFilterBackend, filters.OrderingFilter)
    permission_classes = (permissions.IsAuthenticated,)
    paginator = EnterpriseCustomerSupportPaginator()

    ordering_fields = ["is_admin", "username", "pending_enterprise_customer"]

    def filter_queryset_by_user_query(self, queryset):
        """
        Filter queryset based on user provided query
        """
        user_query = self.request.query_params.get("user_query", None)

        if user_query:
            queryset = queryset.filter(Q(user_email__icontains=user_query) | Q(username__icontains=user_query))
        return queryset

    def retrieve(self, request, *args, **kwargs):
        """
        Filter down the queryset of groups available to the requesting uuid.
        """
        enterprise_uuid = kwargs.get("enterprise_uuid", None)
        ecs_queryset = models.EnterpriseCustomerSupportUsersView.objects.all()

        try:
            ecs_queryset = self.filter_queryset_by_user_query(
                ecs_queryset.filter(enterprise_customer_id=enterprise_uuid)
            )
        except ValidationError:
            # did not find UUID match in either EnterpriseCustomerUser or PendingEnterpriseCustomerUser
            return response.Response(
                {"detail": "Could not find enterprise uuid {}".format(enterprise_uuid)},
                status=status.HTTP_404_NOT_FOUND,
            )

        # default sort criteria
        is_reversed = False
        ordering_criteria = self.request.query_params.get("ordering", None)
        if ordering_criteria:
            is_reversed = ordering_criteria.startswith('-')
            if is_reversed:
                ordering_criteria = ordering_criteria[1:]

        order_by_admin = f"{('-' if is_reversed else '')}is_admin"
        order_by_username_or_email = Case(
            When(is_pending=True, then=F('user_email')),
            default=F('username'),
            output_field=CharField()
        )
        if ordering_criteria in ("administrator", "learner"):
            # Sort by admin status
            ecs_queryset = ecs_queryset.order_by(order_by_admin)
        elif ordering_criteria == "details":
            # If pending user, sort by email, otherwise sort by username
            if is_reversed:
                ecs_queryset = ecs_queryset.order_by(order_by_username_or_email.desc())
            else:
                ecs_queryset = ecs_queryset.order_by(order_by_username_or_email.asc())
        elif not ordering_criteria:
            # Apply default ordering criteria (first by is_admin,
            # then username) only if user does not specify ordering criteria;
            ecs_queryset = ecs_queryset.order_by('-is_admin', order_by_username_or_email.asc())
        users_page = self.paginator.paginate_queryset(ecs_queryset, request, view=self)

        serializer = serializers.EnterpriseUserSerializer(users_page, many=True)
        serializer_data = serializer.data

        return self.paginator.get_paginated_response(serializer_data)
